# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from django.db import models
from django.conf import settings
from django.utils.html import format_html, mark_safe
from etgen import html as xghtml
from lino.utils.html import E, join_elems, tostring, SAFE_EMPTY
from lino.core import constants
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.system.choicelists import DisplayColors
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _, gettext

from lino_prima.lib.prima.roles import PrimaStaff, PrimaTeacher, PrimaPupil

# NOT_RATED = '▨'
# NOT_RATED = '◻' # 25fb white medium square
# NOT_RATED = '□' # 25a1 white square
# NOT_RATED = '▭' # 25ad white rectagle
NOT_RATED = '☐' # 2610 ballot box


class CertificateDetail(dd.DetailLayout):
    main = "ratings verdict"

    ratings = dd.Panel("""
    enrolment period state id
    SectionResponsesByCertificate
    # RatingsByCertificate
    """, _("Ratings"))

    verdict = dd.Panel("""
    social_skills_comment
    final_verdict
    """, _("Verdict"))

class Certificates(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'cert.Certificate'
    column_names = "enrolment period state *"
    detail_layout = "cert.CertificateDetail"
    insert_layout = """
    enrolment
    period
    """

# class MyCertificates(Certificates, My):
#     required_roles = dd.login_required(PrimaTeacher)

class CertificatesByEnrolment(Certificates):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "enrolment"
    default_display_modes = { None: constants.DISPLAY_MODE_SUMMARY}
    # row_template = "{row.period.nickname}"


class CertTemplates(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'cert.CertTemplate'
    detail_layout = """
    designation
    cert.SectionsByTemplate
    """

class CertSections(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'cert.CertSection'
    column_names = "cert_template seqno subject remark *"
    detail_layout = """
    cert_template seqno subject id
    remark
    cert.ElementsBySection
    """

class SectionsByTemplate(CertSections):
    master_key = 'cert_template'
    column_names = "seqno subject remark *"


class CertElements(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'cert.CertElement'
    column_names = "cert_section seqno skill max_score *"
    # detail_layout = """
    # cert_section
    # seqno
    # skill max_score
    # id
    # """

class ElementsBySection(CertElements):
    master_key = 'cert_section'
    column_names = "seqno skill max_score *"


class CertRatings(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'cert.CertRating'

    detail_layout = """
    response response__certificate__enrolment response__certificate__period
    cert_element cert_element__skill max_score
    # ratings_done total_max_score total_score
    computed_rating score rating_buttons
    ratings_report
    """

class RatingsBySkill(CertRatings):
    # master = 'cert.Certificate'
    master_key = 'cert_element__skill'
    required_roles = dd.login_required(PrimaTeacher)
    column_names = "cert_element max_score score *"
    # order_by = ['cert_element__cert_section', 'cert_element']
    default_display_modes = { None: constants.DISPLAY_MODE_SUMMARY}
    row_template = "{row.response.certificate.enrolment}"

class RatingsByResponse(CertRatings):
    master_key = 'response'
    required_roles = dd.login_required(PrimaTeacher)
    column_names = "cert_element #total_score score max_score  *"
    default_display_modes = { None: constants.DISPLAY_MODE_LIST}
    obvious_fields = {'certificate', 'section'}

# class RatingsByCertificate(CertRatings):
#     # master = 'cert.Certificate'
#     master_key = 'response__certificate'
#     required_roles = dd.login_required(PrimaTeacher)
#     column_names = "cert_element total_score max_score score *"
#     order_by = ['cert_element__cert_section', 'cert_element']
#     group_by = [lambda obj: obj.cert_element.cert_section]
#     default_display_modes = { None: constants.DISPLAY_MODE_LIST}
#     # row_template = "{row.cert_element} {row.total_score} / {row.score or '☐'} ({row.ratings_done}% done)"
#
#     @classmethod
#     def before_group_change(cls, gh, obj):
#         return format_html("<h2>{}</h2>", obj.cert_element.cert_section)
#
#     @classmethod
#     def row_as_paragraph(cls, ar, self):
#         text = str(self.cert_element.skill) + " (" + _("computed") + " "
#         text += self.computed_text()
#         # text += " " + str(self.ratings_done) + "% done)"
#         text += "): "
#         # text += ar.obj2htmls(self, str(self.score or NOT_RATED))
#
#         elems = list(map(tostring, self.get_rating_buttons(ar)))
#         text += " | ".join(elems)
#         return mark_safe(text)

from lino.core.roles import Explorer

class SectionResponses(dd.Table):
    required_roles = dd.login_required(Explorer)
    model = "cert.SectionResponse"
    detail_layout = """
    certificate section rating_type max_score score smiley predicate
    remark
    cert.RatingsByResponse
    """

class SectionResponsesByCertificate(SectionResponses):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "certificate"
    column_names = "section total_score total_max_score rating_buttons remark *"



# -*- coding: utf-8 -*-
# fmt: off

import datetime

from atelier.sphinxconf import configure; configure(globals())
from lino.sphinxcontrib import configure; configure(
    globals(), 'lino_prima.projects.prima1.settings')

extensions += ['lino.sphinxcontrib.help_texts_extractor']
help_texts_builder_targets = {'lino_prima.': 'lino_prima.lib.prima'}

project = html_title = "Lino Prima"
copyright = '2016-{} Rumma & Ko Ltd'.format(datetime.date.today().year)

# html_context.update(public_url='https://prima.lino-framework.org')



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from django.db import models
from lino.api import dd, _
from lino_prima.lib.prima.roles import PrimaStaff, PrimaTeacher



class CertificateStates(dd.ChoiceList):
    verbose_name = _("Certificate state")
    verbose_name_plural = _("Certificate states")

add = CertificateStates.add_item
add("1", _("Draft"), "default")  # Unfertig -> Entwurf
add("2", _("To discuss"))  # Zu besprechen
add("3", _("Ready"))       # Fertig



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

# Import a set of csv files that have been created using db2csv.py:
# https://gitlab.com/synodalsoft/zeugnisse/-/blob/main/db2csv.py

# Usage:
# go prima1
# pm initdb std csv2db demo_users demo2 checkdata checksummaries


from pathlib import Path

import datetime

combine = datetime.datetime.combine

import csv
from django.utils.text import format_lazy
from django.conf import settings
from lino.api import rt, dd, _
from lino.utils import Cycler, i2d, ONE_DAY
from lino.utils import AttrDict
from lino.core.roles import SiteAdmin
from lino.modlib.system.choicelists import DurationUnits
from lino_xl.lib.working.roles import Worker
from lino.utils.quantities import Duration
from lino.utils.instantiator import make_if_needed
from lino.utils.mldbc import babel_named as named
from lino.modlib.users.utils import create_user
from lino_xl.lib.working.choicelists import ReportingTypes
# from lino_xl.lib.tickets.choicelists import SiteStates
from lino_xl.lib.tickets.roles import Reporter, TicketsStaff
from lino_xl.lib.tickets.fixtures.demo import TEXTS
from lino.modlib.system.choicelists import DisplayColors


TEXTS = Cycler(TEXTS)
LAST = AttrDict()

UserTypes = rt.models.users.UserTypes
User = rt.models.users.User
Grade = rt.models.prima.Grade
Course = rt.models.prima.Course
Enrolment = rt.models.prima.Enrolment
Challenge = rt.models.ratings.Challenge
Project = rt.models.projects.Project
GeneralRatings = rt.models.projects.GeneralRatings
# ProjectSkill = rt.models.projects.ProjectSkill
FinalExam = rt.models.ratings.FinalExam
Skill = rt.models.prima.Skill
Subject = rt.models.prima.Subject
ProjectSection = rt.models.projects.ProjectSection
StoredPeriod = rt.models.periods.StoredPeriod
Cast = rt.models.prima.Cast
Role = rt.models.prima.Role
Certificate = rt.models.cert.Certificate
CertSection = rt.models.cert.CertSection
CertElement = rt.models.cert.CertElement
CertRating = rt.models.cert.CertRating
SectionResponse = rt.models.cert.SectionResponse

FARBE2COLOR = {
    1:"red",  # Red#ff6d6d,0
    2:"blue",  # Blau,#9797ff,1
    3:"green",  # Grün,#72ff72,2
    4:"yellow",  # Gelb,#ffff70,3
    5:"violet", # Violett,#fc6cfc,4
}


def datedone(s):
    parts = s.split()
    return parts[0]

def uid(old):
    # convert id in Schüler.csv to id in users.User
    return int(old) + 100

def sk2skill(skid):
    return int(skid) + 500


def grade(ref, desig, ct=None):
    kwargs = dd.babelkw('designation', desig)
    kwargs.update(ref=str(ref))
    kwargs.update(cert_template_id=ct)
    return Grade(**kwargs)

def update_or_create(m, **kwargs):
    if (obj := m.objects.filter(id=kwargs['id']).first()) is not None:
        for k, v in kwargs.items():
            setattr(obj, k, v)
        return obj
    return m(**kwargs)

def get_or_create(model, **kwargs):
    # similar to Djanop's QuerySet.get_or_create() but calls full_clean
    try:
        obj = model.objects.get(**kwargs)
    except model.DoesNotExist:
        obj = model(**kwargs)
        obj.full_clean()
        obj.save()
    return obj

def find_teachers(s):
    s = s.lower()
    names = {s}
    for sep in (",", " und "):
        if sep in s:
            names = set([i.strip() for i in s.split(sep)])
    for u in User.objects.all():
        if f"{u.first_name} {u.last_name}".lower() in names:
            yield u
        elif f"{u.last_name} {u.first_name}".lower() in names:
            yield u

class Loader:
    def __init__(self):
        self.max_exam_id = 0
        self.max_group_id = 0

    def common_objects(self):

        STUFEN = (
            (1, 'Erste Klasse'),
            (3, 'Zweite Klasse'),
            (4, 'Mittelstufe'),
            (2, 'Oberstufe'))
        for k, v in STUFEN:
            yield rt.models.cert.CertTemplate(id=k, designation=v)

        yield grade(1, _("First grade"), 1)
        yield grade(2, _("Second grade"), 3)
        yield grade(3, _("Third grade"), 4)
        yield grade(4, _("Fourth grade"), 4)
        yield grade(5, _("Fifth grade"), 2)
        yield grade("6", _("Sixth grade"), 2)
        yield grade("X", _("Alumni"))

        self.SOZIALVERHALTEN = Subject(id=20, designation="Sozialverhalten", rating_type="smiley")
        yield self.SOZIALVERHALTEN
        self.ARBEITSVERHALTEN = Subject(id=21, designation="Arbeitsverhalten", rating_type="smiley")
        yield self.ARBEITSVERHALTEN

        self.ska2section = dict()
        # SozialKompetenzAbschnitt.csv
        for (id,vorlage,titel,ord_outer) in [
            (2,2,self.SOZIALVERHALTEN,0),
            (15,3,self.SOZIALVERHALTEN,0),
            (17,4,self.SOZIALVERHALTEN,0),
            (13,1,self.SOZIALVERHALTEN,0),
            (18,4,self.ARBEITSVERHALTEN,1),
            (1,2,self.ARBEITSVERHALTEN,1),
            (14,1,self.ARBEITSVERHALTEN,1),
            (16,3,self.ARBEITSVERHALTEN,1)
        ]:
            yield (obj := CertSection(
                id=id+100, cert_template_id=vorlage, subject=titel,
                seqno=ord_outer+100))
            self.ska2section[id] = obj

        self.klassenleiter = Role(designation="Klassenleiter")
        yield self.klassenleiter
        self.geographielehrerin = Role(designation="Geographielehrerin")
        yield self.geographielehrerin
        self.geschichtslehrerin = Role(designation="Geschichtslehrerin")
        yield self.geschichtslehrerin
        self.techniklehrerin = Role(designation="Techniklehrerin")
        yield self.techniklehrerin
        self.ev_religionslehrerin = Role(designation="Evangelisch")
        yield self.ev_religionslehrerin
        self.ka_religionslehrerin = Role(designation="Katholisch")
        yield self.ka_religionslehrerin
        self.bausteine = Role(designation="Bausteine")
        yield self.bausteine

    def make_cast_if_needed(self, subject, teacher, group, role):
        yield make_if_needed(Course, subject=subject, group=group)

        # Make sure that there is at least one cast for this teacher and this
        # group. Don't create more than one cast per teacher and group. Let it
        # be if there is already another cast, anyway we are just guessing the
        # casts:
        qs = Cast.objects.filter(user=teacher, group=group)
        if not qs.exists():
            yield make_if_needed(Cast, role=role, user=teacher, group=group)



def csv2db(loader, import_dir, start_year):

    import_dir = Path(import_dir).expanduser()

    def csv2dicts(filename):
        filename = import_dir / filename
        print(f"Reading {filename}")
        with open(filename) as csvfile:
            for r in csv.DictReader(csvfile):
                yield r

    # unknown_teacher = User.objects.order_by("id").first()
    # # unknown_teacher = User.objects.get(username="robin")
    # print(f"The unknown teacher is {unknown_teacher}")

    periode1 = StoredPeriod.get_or_create_from_date(datetime.date(start_year, 9, 1))
    periode2 = StoredPeriod.get_or_create_from_date(datetime.date(start_year+1, 3, 1))

    if periode1.year != periode2.year:
        raise Exception(f"{periode1.year} is not {periode2.year}")
    print(f"Importing {periode1.year}")

    def p2p(s):
        if s == "1":
            return periode1
        elif s == "2":
            return periode2
        else:
            raise Exception(f"20241023 Unknown period {repr(s)}")
        # periode
        # return StoredPeriod.objects.get(id=int(old))

    def klasse2group(i):
        i = int(i)
        if periode1.year.ref == "2024/25":
            i += 1000
        loader.max_group_id = max(loader.max_group_id, i)
        return i

    def examid(i):
        i = int(i)
        if periode1.year.ref == "2024/25":
            i += 10000  # 20241022 2023/24 max exam id was 9162
        loader.max_exam_id = max(loader.max_exam_id, i)
        return i

    def sectionresponse_id(i):
        i = int(i)
        if periode1.year.ref == "2024/25":
            i += 55000  # 20241128 2023/24 max SectionResponse id was 52066
        return i

    def get_certificate(r):
        p = p2p(r['periode'])
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']),
                group__year=p.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! Enrolment does not exist: {r}")
            # continue
        try:
            return Certificate.objects.get(
                enrolment=enr,
                period=p)
        except Certificate.DoesNotExist:
            raise Exception(f"Oops! Certificate does not exist: {enr} {p}")
            # continue

    main_subjects = []
    science_subject = None
    sport_subject = None
    religion_subject = None
    for r in csv2dicts("Fach.csv"):
        # id,name,advanced
        kwargs = dict(id=r['id'], designation=r['name'], advanced=r['advanced'])
        if r['name'] in {"Musik", "Kunst", "Sport"}:
            kwargs.update(rating_type="predicate")
        yield (obj := update_or_create(Subject, **kwargs))
        if obj.designation == ("Sport"):
            sport_subject = obj
        if obj.designation == ("Religion"):
            religion_subject = obj
        if obj.designation == ("Wissenschaften"):
            science_subject = obj
        if obj.designation in {"Deutsch", "Französisch", "Mathematik"}:
            main_subjects.append(obj)

    for r in csv2dicts("Kompetenz.csv"):
        # id,beschreibung,fach,mit_abschlusstests
        subject = Subject.objects.get(id=r['fach'])
        if subject == science_subject:
            try:
                subject = Subject.objects.get(designation=r['beschreibung'])
            except Subject.DoesNotExist:
                subject = Subject(designation=r['beschreibung'])
                yield subject
                # print(f"20241121 Created science subject {subject}")
        kwargs = dict(
            id=r['id'],
            designation=r['beschreibung'],
            subject=subject,
            with_exams=r['mit_abschlusstests']
        )
        yield update_or_create(Skill, **kwargs)

    for r in csv2dicts("Baustein.csv"):
        # id,name,kürzel,farbe,jahrgang,hauptkompetenz
        kwargs = dict(
            id=r['id'], designation=r['name'],
            short_header=r['kürzel'],
            display_color=FARBE2COLOR[int(r['farbe'])],
            main_skill_id=r['hauptkompetenz'],
            grade=Grade.objects.get(id=r['jahrgang'])
        )
        yield update_or_create(rt.models.projects.ProjectTemplate, **kwargs)

    for r in csv2dicts("BausteinAbschnitt.csv"):
        # id,name,leistungsblock_ptr,baustein,ord_outer
        kwargs = dict(
            id=r['id'], designation=r['name'],
            project_template_id=r['baustein'],
            seqno=int(r['ord_outer'])+1
        )
        yield update_or_create(ProjectSection, **kwargs)

    # for r in csv2dicts("BausteinArbeit.csv"):
    #     # id,name,max_punkte
    #     kwargs = dict(
    #         id=r['id'], designation=r['name'],
    #         max_score=r['max_punkte']
    #     )
    #     yield update_or_create(rt.models.projects.ProjectSkill, **kwargs)

    for r in csv2dicts("ZeugnisAbschnitt.csv"):
        # id,vorlage,fach,fixtext,ord_outer
        # subject = Subject.objects.get(id=r['fach'])
        kwargs = dict(
            id=r['id'],
            cert_template_id=r['vorlage'],
            # subject=subject,
            subject_id=r['fach'],
            remark=r['fixtext'],
            # advanced=subject.advanced,
            seqno=int(r['ord_outer'])+1,
        )
        yield update_or_create(CertSection, **kwargs)

    for r in csv2dicts("SozialKompetenz.csv"):
        # id,abschnitt,name,ord_inner
        cs = loader.ska2section[int(r['abschnitt'])]
        kwargs = dict(
            id=sk2skill(r['id']),
            subject=cs.subject,
            designation=r['name'],
            with_exams=False,
            # with_smilies=True,
        )
        # seqno=r['ord_inner']
        yield update_or_create(Skill, **kwargs)

    for r in csv2dicts("ZeugnisElement.csv"):
        # id,abschnitt,kompetenz,max_punkte,ord_inner
        kwargs = dict(
            id=r['id'],
            cert_section_id=r['abschnitt'],
            skill_id=r['kompetenz'],
            max_score=r['max_punkte'],
            seqno=int(r['ord_inner'])+1,
        )
        yield update_or_create(CertElement, **kwargs)

    for r in csv2dicts("Teacher.csv"):
        # id,password,last_login,is_superuser,username,first_name,last_name,email,is_staff,is_active,date_joined
        kwargs = dict(
            id=r['id'],
            username=r['username'],
            first_name=r['first_name'],
            last_name=r['last_name'],
            email=r['email'],
            password=r['password'],
        )
        if r['is_superuser'] == "True":
            kwargs.update(user_type=UserTypes.admin)
        else:
            kwargs.update(user_type=UserTypes.teacher)
        yield update_or_create(User, **kwargs)

    for r in csv2dicts("Klasse.csv"):
        # id, name, jahrgang, klassenleiterinnen, ev_religionslehrerin,
        # ka_religionslehrerin, sportlehrerin, geographielehrerin,
        # geschichtslehrerin, techniklehrerin
        kwargs = dict(
            id=klasse2group(r['id']),
            designation="{jahrgang}{name}".format(**r),
            year=periode1.year,
            grade=Grade.objects.get(id=r['jahrgang'])
        )
        yield (grp := rt.models.prima.Group(**kwargs))

        if False:
            for t in find_teachers(r['klassenleiterinnen']):
                for s in main_subjects:
                    yield Cast(user=t, group=grp, subject=s, role=loader.klassenleiter)
            if religion_subject is not None:
                for t in find_teachers(r['ev_religionslehrerin']):
                    yield Cast(user=t, group=grp, subject=religion_subject,
                        role=loader.ev_religionslehrerin)
                for t in find_teachers(r['ka_religionslehrerin']):
                    yield Cast(user=t, group=grp, subject=religion_subject,
                        role=loader.ka_religionslehrerin)
            if sport_subject is not None:
                for t in find_teachers(r['sportlehrerin']):
                    yield Cast(user=t, group=grp, subject=sport_subject)
            if science_subject is not None:
                for t in find_teachers(r['geographielehrerin']):
                    yield Cast(user=t, group=grp, subject=science_subject,
                        role=loader.geographielehrerin)
                for t in find_teachers(r['geschichtslehrerin']):
                    yield Cast(user=t, group=grp, subject=science_subject,
                        role=loader.geschichtslehrerin)
                for t in find_teachers(r['techniklehrerin']):
                    yield Cast(user=t, group=grp, subject=science_subject,
                        role=loader.techniklehrerin)

    print(f"20241022 {periode1.year.ref} max group id was {loader.max_group_id}")

    # Seems that this model isn't used anywhere
    # for r in csv2dicts("LehrerRolle.csv"):
    #     # id,lehrer,klasse,fach
    #     kwargs = dict(
    #         # id=r['id'],
    #         user_id=r['lehrer'],
    #         group_id=klasse2group(r['klasse']),
    #         subject_id=r['fach']
    #     )
    #     yield Cast(**kwargs)

    for r in csv2dicts("Schüler.csv"):
    # id,name,vorname,klasse
        yield update_or_create(User,
            id=uid(r['id']),
            username="Schüler#" + r['id'],
            last_name=r['name'],
            first_name=r['vorname'],
            nickname=r['vorname'],
        )
        pupil_id = uid(r['id'])
        obj = Enrolment(
            # id=r['id'],
            pupil_id=pupil_id,
            group_id=klasse2group(r['klasse'])
        )
        yield obj
        # if pupil_id == 426:
        #     print(f"20241020 created enrolment for {pupil_id} in {obj.group.year}")


    for r in csv2dicts("AbschlussTest.csv"):
        # id,name,kompetenz,max_punkte
        kwargs = dict(
            id=r['id'],
            designation=r['name'],
            skill_id=r['kompetenz'],
            max_score=r['max_punkte'])
        yield update_or_create(FinalExam, **kwargs)

    for r in csv2dicts("AbschlussTestBewertung.csv"):
        # id,polymorphic_ctype,datum,periode,lehrer,schüler,bewertung_ptr,test,punkte
        p = p2p(r['periode'])
        # if p != periode1:
        #     print(f"20241107 {p} is not {periode1} in AbschlussTestBewertung {r}")
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']),
                group__year=p.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! Enrolment does not exist: {r}")
            # print(f"Oops! Enrolment does not exist: {r}")
            continue

        exam = FinalExam.objects.get(id=r['test'])
        teacher = User.objects.get(id=r['lehrer'])
        kwargs = dict(
            id=r['id'],
            enrolment=enr,
            date_done=datedone(r['datum']),
            teacher=teacher,
            exam=exam,
            period=p,
            score=r['punkte'])
        yield update_or_create(rt.models.ratings.FinalExamRating, **kwargs)
        yield loader.make_cast_if_needed(exam.skill.subject, teacher, enr.group, loader.klassenleiter)

    for r in csv2dicts("Test.csv"):
    # id,name,leistungsblock_ptr,klasse,periode,fach
        kwargs = dict(
                group_id=klasse2group(r['klasse']),
                # user=unknown_teacher,
                subject_id=r['fach'])
        # qs = Cast.objects.filter(**kwargs)
        # if qs.count() == 0:
        #     cast = Cast(user=unknown_teacher, **kwargs)
        #     yield cast
        # elif qs.count() == 1:
        #     cast = qs.first()
        # else:
        #     cast = qs.first()
        #     # no way to know which teacher did this test, so we take the first
        #     # raise Exception(f"Multiple casts for {kwargs}!")
        kwargs.update(
            id=examid(r['id']), heading=r['name'],
            period=p2p(r['periode']),
            # cast=cast
        )
        yield rt.models.ratings.Exam(**kwargs)
    print(f"20241022 {periode1.year.ref} max exam id was {loader.max_exam_id}")

    dup_block = []
    no_block = []
    mult_block = []
    for r in csv2dicts("Leistung.csv"):
        # id,block,kompetenz,max_punkte,ord_inner
        # block points to either an Exam or a ProjectSection
        kwargs = dict(
            id=r['id'],
            # exam_type='test',
            skill_id=r['kompetenz'],
            seqno=int(r['ord_inner']) + 1,
            # seqno=f"{r['block']}{int(r['ord_inner']):03d}",
            max_score=r['max_punkte']
        )
        block = examid(r['block'])
        if rt.models.ratings.Exam.objects.filter(id=block).exists():
            kwargs.update(exam_id=block)
            yield update_or_create(Challenge, **kwargs)
            # if rt.models.ratings.ProjectSection.objects.filter(pk=block).exists():
            #     dup_block.append(r['id'])
            #     raise Exception(f"Found both a Exam and a ProjectSection for {r}")
            # else:
            #     kwargs.update(exam_id=block)
            #     yield Challenge(**kwargs)
        else:
            block = int(r['block'])
            qs = ProjectSection.objects.filter(pk=block)
            n = qs.count()
            if n == 1:
                # print(f"Ignore {r}")
                kwargs.update(project_section=qs.first())
                yield update_or_create(Challenge, **kwargs)
            elif n == 0:
                no_block.append(r['id'])
            else:
                raise Exception(f"20241022 multiple project sections for {r}")
                mult_block.append(r['id'])
            # prjskills = rt.models.ratings.ProjectSkill.objects.all()
            # kwargs.update(project_section_id=block)
            # for i in prjskills:
            #     kwargs.update(project_skill=i)
            #     yield rt.models.ratings.Challenge(**kwargs)
    # print(f"{len(dup_block)} challenges having both an Exam and a ProjectSection: {dup_block}")
    print(f"Ignored {len(no_block)} challenges having neither Exam nor ProjectSection")
    # print(f"{len(mult_block)} challenges having multiple ProjectSection: {mult_block}")

    # No need to import LeistungsBlock.csv because that data is in Test.csv and
    # BlockAbschnitt.csv:
    # for r in csv2dicts("BlockAbschnitt.csv"):

    for r in csv2dicts("LeistungsBewertung.csv"):
    # id,polymorphic_ctype,datum,periode,lehrer,schüler,bewertung_ptr,leistung,punkte
        p = p2p(r['periode'])
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']),
                group__year=p.year)
        except Enrolment.DoesNotExist:
            raise
            # print(f"Oops! Enrolment does not exist: {r}")
            # continue
        challenge = Challenge.objects.get(id=r['leistung'])
        teacher = User.objects.get(id=r['lehrer'])
        kwargs = dict(
            id=r['id'],
            enrolment=enr,
            date_done=datedone(r['datum']),
            teacher=teacher,
            challenge=challenge,
            period=p,
            score=r['punkte'])
        yield update_or_create(rt.models.ratings.ChallengeRating, **kwargs)
        if challenge.exam_id is None:
            trole = loader.bausteine
        else:
            trole = loader.klassenleiter
            if challenge.exam.user_id is None:
                challenge.exam.user = teacher
                challenge.exam.full_clean()
                challenge.exam.save()
        yield loader.make_cast_if_needed(challenge.skill.subject, teacher, enr.group, trole)

    for r in csv2dicts("BausteinArbeitBewertung.csv"):
    # id,polymorphic_ctype,datum,periode,lehrer,schüler,bewertung_ptr,arbeit,baustein,punkte
        p = p2p(r['periode'])
        pupil_id = uid(r['schüler'])
        try:
            enr = Enrolment.objects.get(
                pupil_id=pupil_id,
                group__year=p.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No enrolment for {pupil_id} in {p.year}") from None
            continue
        prj, created = Project.objects.get_or_create(
            enrolment=enr, template_id=r['baustein'])

        gr = GeneralRatings.get_by_value(r['arbeit'])
        setattr(prj, gr.field_name, r['punkte'])
        yield prj

    for r in csv2dicts("BausteinKommentar.csv"):
        if not r['kommentar'].strip():
            continue
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']), group__year=periode1.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No enrolment for BausteinKommentar {r}")
            continue
        prj, created = Project.objects.get_or_create(
            enrolment=enr, template_id=r['baustein'])
        prj.remark = r['kommentar']
        yield prj

    for r in csv2dicts("Zeugnis.csv"):
        # id,schüler,periode,status,katholische_religion,sozialverhalten_kommentar,versetzungsentscheidung,abwesenheiten_mit_entschuldigung,abwesenheiten_mit_attest,abwesenheiten_ohne,förderziele
        p = p2p(r['periode'])
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']), group__year=periode1.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No enrolment for Zeugnis {r}")
            continue
        kwargs = dict(
            id=r['id'],
            enrolment=enr,
            period=p,
            social_skills_comment=r['sozialverhalten_kommentar'],
            final_verdict=r['versetzungsentscheidung']
        )
        yield Certificate(**kwargs)

    for r in csv2dicts("TestKommentar.csv"):
        # id,schüler,test,kommentar
        try:
            enr = Enrolment.objects.get(
                pupil_id=uid(r['schüler']), group__year=periode1.year)
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No enrolment for Zeugnis {r}")
            continue
        kwargs = dict(
            id=r['id'],
            enrolment=enr,
            exam_id=examid(r['test']),
            remark=r['kommentar'],
        )
        yield rt.models.ratings.ExamResponse(**kwargs)

    max_response_id = 0
    for r in csv2dicts("ZeugnisFachKommentar.csv"):
        # id,schüler,periode,abschnitt,kommentar
        cert = get_certificate(r)
        # p = p2p(r['periode'])
        # try:
        #     enr = Enrolment.objects.get(
        #         pupil_id=uid(r['schüler']), group__year=periode1.year)
        # except Enrolment.DoesNotExist:
        #     raise Exception(f"Oops! No enrolment for Zeugnis {r}")
        try:
            cs = CertSection.objects.get(id=r['abschnitt'])
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No enrolment for Zeugnis {r}")
        kwargs = dict(
            # id=sectionresponse_id(r['id']),
            certificate=cert,
            section=cs,
            # enrolment=enr,
            # period=p,
            # subject=cs.subject,
            remark=r['kommentar'],
        )
        response = SectionResponse(**kwargs)
        yield response
        max_response_id = max(response.id, max_response_id)

    print(f"20241128 {periode1.year.ref} max SectionResponse id was {max_response_id}")

    for r in csv2dicts("ZeugnisPrädikatNote.csv"):
        # id,schüler,periode,abschnitt,note
        cert = get_certificate(r)
        try:
            cs = CertSection.objects.get(id=r['abschnitt'])
        except Enrolment.DoesNotExist:
            raise Exception(f"Oops! No CertSection for {r}")
        resp, created = SectionResponse.objects.get_or_create(
            certificate=cert, section=cs)
        resp.score = r['note']
        yield resp
        # kwargs = dict(
        #     id=r['id'],  # the id isn't used anywhere, so we can safely let Lino assign new id's
        #     response=resp,
        #     score=r['note'],
        # )
        # yield SectionResponse(**kwargs)


    for r in csv2dicts("ZeugnisKompetenzNote.csv"):
        # id,schüler,periode,element,note
        cert = get_certificate(r)
        elem = CertElement.objects.get(id=r['element'])
        resp, created = SectionResponse.objects.get_or_create(
            certificate=cert, section=elem.cert_section)
        kwargs = dict(
            # id=r['id'],  # the id isn't used anywhere, so we can safely let Lino assign new id's
            response=resp,
            cert_element=elem,
            # certificate=cert,
            # enrolment=enr,
            # period=p,
            score=r['note'],
        )
        yield CertRating(**kwargs)

    # ignored = []
    for r in csv2dicts("ZeugnisSozialNote.csv"):
        # id,schüler,periode,sozialkompetenz,note
        cert = get_certificate(r)
        tpl = cert.enrolment.group.grade.cert_template
        try:
            skill = Skill.objects.get(pk=sk2skill(r['sozialkompetenz']))
        except Skill.DoesNotExist:
            raise Exception(f"Oops! Skill does not exist: {r}")
        cs, created = CertSection.objects.get_or_create(
            cert_template=tpl, subject=skill.subject)
        resp = get_or_create(SectionResponse, certificate=cert, section=cs)
        elem = get_or_create(CertElement, skill=skill, cert_section=cs)
        kwargs = dict(
            # id=r['id'],  # the id isn't used anywhere, so we can safely let Lino assign new id's
            # certificate=cert,
            response=resp,
            cert_element=elem,
            score=r['note'],
        )
        yield CertRating(**kwargs)

    # if len(ignored) > 0:
    #     print(f"Ignored {len(ignored)} ratings because CertElement does not exist.")


def objects():

    if False:
        settings.SITE.site_config.update(simulate_today=i2d(20240601))

    loader = Loader()
    yield loader.common_objects()
    # yield csv2db("~/work/zeugnisse/export", 2023)
    yield csv2db(loader, "~/Downloads/export/2023", 2023)
    yield csv2db(loader, "~/Downloads/export/2024", 2024)

    # print("compute_project_sums()")
    # rt.models.ratings.compute_project_sums()



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from django.db import models
from django.db.models import Q
from django.conf import settings
from django.utils.html import format_html, mark_safe
from etgen import html as xghtml
from lino.utils.html import E, join_elems, tostring, SAFE_EMPTY
from lino.core import constants
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.system.choicelists import DisplayColors
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _, gettext

from lino_prima.lib.prima.roles import PrimaStaff, PrimaTeacher, PrimaPupil
from lino_prima.lib.ratings.mixins import NOT_RATED, format_score

from .choicelists import GeneralRatings


# class ProjectSkills(dd.Table):
#     required_roles = dd.login_required(PrimaStaff)
#     model = 'projects.ProjectSkill'
#

class ProjectTemplates(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'projects.ProjectTemplate'
    column_names = "grade short_header designation *"
    detail_layout = """
    designation short_header
    display_color id
    main_skill grade
    SectionsByProject
    """

class ProjectTemplatesBySkill(ProjectTemplates):
    master_key = "main_skill"

class ProjectTemplatesByGrade(ProjectTemplates):
    master_key = "grade"


class ProjectSections(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'projects.ProjectSection'
    detail_layout = """
    seqno designation
    project_template id
    ratings.ChallengesByProjectSection
    """
    default_display_modes = {None: constants.DISPLAY_MODE_LIST}

class SectionsByProject(ProjectSections):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "project_template"

class ProjectDetail(dd.DetailLayout):
    main = "general more"

    general = dd.Panel("""
    enrolment date_started
    remark
    general_ratings:30 ratings.ChallengeRatingsByProject:40
    """, _("General"))

    more = dd.Panel("""
    template id
    ratings_done total_score total_max_score
    # ratings.RatingsByProject
    """, _("More"))

    general_ratings = "\n".join(
        [pf.field_name for pf in GeneralRatings.get_list_items()])

class Projects(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'projects.Project'
    detail_layout = "projects.ProjectDetail"


class ProjectsByTemplate(Projects):
    master_key = "template"

class ProjectsByEnrolment(Projects):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "enrolment"
    default_display_modes = { None: constants.DISPLAY_MODE_SUMMARY}
    insert_layout = dd.InsertLayout("""
    enrolment
    template
    remark
    """, window_size=(40, 15)) # , hidden_elements="template")


# class ProjectRatings(dd.Table):
#     required_roles = dd.login_required(PrimaTeacher)
#     model = 'ratings.ProjectRating'
#     detail_layout = dd.DetailLayout("""
#     project_skill score project_skill__max_score
#     project
#     teacher period date_done
#     """, window_size=(60, "auto"))
#
# class AllProjectRatings(ProjectRatings):
#     required_roles = dd.login_required(PrimaStaff)

# class RatingsByProject(ProjectRatings):
#     master_key = "project"
#     label = _("Project ratings")
#     default_display_modes = { None: constants.DISPLAY_MODE_LIST}


from lino_prima.lib.prima.ui import EnrolmentsByGroup

# class PupilsAndProjectsByGroup(dd.VirtualTable):
class PupilsAndProjectsByGroup(EnrolmentsByGroup):
    required_roles = dd.login_required(PrimaTeacher)
    master = "prima.Group"
    label = _("Projects")
    # label = _("Projects by group")
    default_display_modes = {None: constants.DISPLAY_MODE_HTML}

    @classmethod
    def table_as_html(cls, ar):
        ProjectTemplate = rt.models.projects.ProjectTemplate
        Enrolment = rt.models.prima.Enrolment
        Project = rt.models.projects.Project
        grp = ar.master_instance
        # print(f"20241017 cellattrs {ar.renderer.cellattrs}")
        templates = ProjectTemplate.objects.filter(grade=grp.grade).order_by("display_color")
        if not templates.exists():
            return _("There are no {projects} configured for grade {grade}.").format(
                projects=Project._meta.verbose_name_plural, grade=grp.grade)

        cellstyle = "padding:2pt; margin:0pt; text-align:center;"
        insert_button_attrs = dict(style="text-align:center;")
        def makecell(sar, tpl):
            try:
                prj = Project.objects.get(
                    enrolment=sar.master_instance, template=tpl)
            except Project.DoesNotExist:
                btn = sar.gen_insert_button(None, insert_button_attrs,
                        template=str(tpl), templateHidden=tpl.pk)
                if btn is None:
                    return ""
                return E.td(btn, style=cellstyle)
                # return E.p(txt, btn, align="center")
                # return str("+")
            if prj.ratings_done == 100:
                color = "#48c78e"
                txt = "☑" # U+2611
            else:
                color = "#ffe08a"
                txt = "⚒" # U+2692
            if False:
                if prj.total_max_score:
                    score = format_score(100 * prj.total_score / prj.total_max_score) + "%"
                else:
                    score = NOT_RATED
                if prj.ratings_done is None:
                    cv = 0
                else:
                    cv = int(prj.ratings_done / 20)  # a value 0..5
                # completion = "▇" * cv + "▁" * (5-cv)
                # completion = "|" * cv + "." * (5-cv)
                completion = "▮" * cv + "▯" * (5-cv)
                txt = f"{completion} {score}"
            return E.td(sar.obj2html(prj, txt),
                style=cellstyle + "background-color:" + color)

        table = xghtml.Table()
        table.attrib.update(ar.renderer.tableattrs)
        headers = [E.td(gettext("Pupil"))]
        # cellstyle = "text-align:center;"
        for prj in templates:
            # print(f"20241017 {prj.display_color}")
            headers.append(E.td(
                ar.obj2html(prj, prj.short_header,
                    style=f"color:{prj.display_color.font_color};"),
                style=cellstyle + f"background-color:{prj.display_color.name};"))
        table.head.append(E.tr(*headers))
        for enr in Enrolment.objects.filter(group=grp):
            sar = ProjectsByEnrolment.request(parent=ar, master_instance=enr)
            cells = [E.td(ar.obj2html(enr, str(enr.pupil)))]
            for prj in templates:
                cells.append(makecell(sar, prj))
            table.body.append(E.tr(*cells))

        el = table.as_element()
        # if len(toolbar := ar.plain_toolbar_buttons()):
        #     el = E.div(el, E.p(*toolbar))
        return el



# how to run a single test:
# $ python -m unittest tests.test_demo.Main.test_prima1

from lino.utils.pythontest import TestCase


class Main(TestCase):

    demo_projects_root = "lino_prima/projects"

    def test_prima1(self):
        self.do_test_demo_project('prima1')



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from lino.ad import Plugin, _


class Plugin(Plugin):

    verbose_name = _("Projects")
    # menu_group = "ratings"

    def setup_config_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('projects.ProjectTemplates')

    def setup_explorer_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('projects.ProjectSections')
        m.add_action('projects.GeneralRatings')



# -*- coding: utf-8 -*-
# Copyright 2011-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
You can run only these tests by issuing::

  $ python setup.py test -s tests.DemoTests.test_prima1

Or::

  $ go prima
  $ cd lino_prima/projects/prima1
  $ python manage.py test tests.test_simple

"""

from django.core.exceptions import ValidationError
from lino.utils.djangotest import RemoteAuthTestCase
from lino.utils.instantiator import create_row
from lino.api import rt


class SimpleTests(RemoteAuthTestCase):
    maxDiff = None

    def test01(self):
        User = rt.models.users.User
        UserTypes = rt.models.users.UserTypes

        robin = create_row(User,
                           username='rolf',
                           user_type=UserTypes.admin,
                           language="de")



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from django.utils.text import format_lazy
from lino.api import dd, _


class Smilies(dd.ChoiceList):
    verbose_name = _("Smiley")
    verbose_name_plural = _("Smilies")

add = Smilies.add_item
add('4', _("++"), button_text="😁")  # 1F601
add('3', _("+"), button_text="😀")  # 1F600
add('2', _("-"), "default", button_text="😐")  # 1F610)
add('1', _("--"), button_text="😢")  # 1F622
# add('1', _("sad"), button_text="😒")  # 1F612
# add('0', _("very sad"), button_text="😢")  # 1F622


class Predicates(dd.ChoiceList):
    verbose_name = _("Predicate")
    verbose_name_plural = _("Predicates")

add = Predicates.add_item
add('1', _("very good"))
add('2', _("good"))
add('3', _("sufficient"), "default")  # ausreichend
add('4', _("deficient"))  # mangelhaft
add('5', _("insufficient"))  # ungenügend


class RatingType(dd.Choice):
    def __init__(self, value, text, rating_choicelist):
        self.rating_choicelist = rating_choicelist
        self.field_name = value
        super().__init__(value, text, value)

    # def get_rating_choices(self):
    #     if self.rating_choicelist is not None:
    #         return self.rating_choicelist.get_list_items()
    #     return None

class RatingTypes(dd.ChoiceList):
    item_class = RatingType
    verbose_name = _("Rating type")
    verbose_name_plural = _("Rating types")

add = RatingTypes.add_item
# add("10", _("Default"), None, 'default')
add("smiley", _("Smilies"), Smilies)
add("predicate", _("Predicates"), Predicates)


class GeneralRating(dd.Choice):

    def __init__(self, value, text, max_score):
        self.field_name = 'gr_' + value
        self.max_score = max_score
        super().__init__(value, text)

    def get_field(self):
        verbose_name = format_lazy(_("{} ({} points)"), self.text, self.max_score)
        return dd.DecimalField(verbose_name, max_digits=4, decimal_places=1, blank=True, null=True)


class GeneralRatings(dd.ChoiceList):
    item_class = GeneralRating
    verbose_name = _("General rating")
    verbose_name_plural = _("General ratings")


add = GeneralRatings.add_item
add("1", _("Cleanliness"), 3)  # Sauberkeit
add("2", _("Correction"), 3)  # Korrektur
add("3", _("Time managment"), 3)  # Zeitmanagement
add("4", _("Work behaviour"), 3)  # Arbeitsverhalten


@dd.receiver(dd.pre_analyze)
def inject_general_rating_fields(sender, **kw):
    for pf in GeneralRatings.get_list_items():
        dd.inject_field('ratings.Project', pf.field_name, pf.get_field())



# -*- coding: UTF-8 -*-
# Copyright 2024-2025 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from django.db import models
from django.db.models import Q
from django.conf import settings
from django.utils.html import format_html, mark_safe
from etgen import html as xghtml
from lino.utils.html import E, join_elems, tostring, SAFE_EMPTY
from lino.core import constants
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.system.choicelists import DisplayColors
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _, gettext

from lino_prima.lib.prima.roles import PrimaStaff, PrimaTeacher, PrimaPupil
from lino.core.roles import Explorer

from .choicelists import GeneralRatings
from .mixins import NOT_RATED, format_score


class RatingsSummaries(dd.Table):
    required_roles = dd.login_required(Explorer)
    model = "ratings.RatingsSummary"
    extra_display_modes = { constants.DISPLAY_MODE_LIST }
    detail_layout = """
    id master skill enrolment challenge
    ratings.ChallengesByGroupSkill:70 ratings.EnrolmentsByGroupSkill:70
    """

class SummariesByCourse(RatingsSummaries):
    required_roles = dd.login_required(PrimaTeacher)
    master = "prima.Course"
    column_names = "skill score max_score *"

    @classmethod
    def get_request_queryset(self, ar, **filter):
        mi = ar.master_instance
        assert isinstance(mi, rt.models.prima.Course)
        qs = super().get_request_queryset(ar, **filter)
        qs = qs.filter(
            master=mi.group, skill__subject=mi.subject,
            enrolment__isnull=True, challenge__isnull=True)
        return qs


class SkillsByGroup(RatingsSummaries):
    required_roles = dd.login_required(PrimaTeacher)
    # master = "prima.Group"
    master_key = "master"
    filter = Q(challenge__isnull=True, enrolment__isnull=True)
    label = _("Skills overview")
    column_names = "skill__subject skill score *"
    group_by = [lambda obj: obj.skill.subject]
    default_display_modes = {
        70: constants.DISPLAY_MODE_HTML,
        None: constants.DISPLAY_MODE_LIST}
    # default_display_modes = {None: constants.DISPLAY_MODE_HTML}

    @classmethod
    def before_group_change(cls, gh, obj):
        return format_html("<h2>{}</h2>", obj.skill.subject)

class ChallengesByGroupSkill(RatingsSummaries):
    required_roles = dd.login_required(PrimaTeacher)
    master = "ratings.RatingsSummary"
    # master_key = "master"
    filter = Q(challenge__isnull=False, enrolment__isnull=True)
    label = _("Challenge ratings")
    # default_display_modes = {None: constants.DISPLAY_MODE_HTML}
    default_display_modes = {None: constants.DISPLAY_MODE_GRID}
    column_names = "challenge__exam_or_section score *"

    @classmethod
    def setup_request(self, ar):
        super().setup_request(ar)
        mi = ar.master_instance
        ar.known_values.update(master=mi.master, skill=mi.skill)

    # @classmethod
    # def get_request_queryset(self, ar, **filter):
    #     mi = ar.master_instance
    #     assert isinstance(mi, rt.models.ratings.RatingsSummary)
    #     qs = super().get_request_queryset(ar, **filter)
    #     qs = qs.filter(master=mi.master, skill=mi.skill)
    #     return qs

class EnrolmentsByGroupSkill(RatingsSummaries):
    required_roles = dd.login_required(PrimaTeacher)
    master = "ratings.RatingsSummary"
    # master = "prima.Group"
    # master_key = "master"
    filter = Q(challenge__isnull=True, enrolment__isnull=False)
    label = _("Pupil ratings")
    default_display_modes = {None: constants.DISPLAY_MODE_GRID}
    column_names = "enrolment score *"

    @classmethod
    def setup_request(self, ar):
        super().setup_request(ar)
        mi = ar.master_instance
        ar.known_values.update(master=mi.master, skill=mi.skill)

    # @classmethod
    # def get_request_queryset(self, ar, **filter):
    #     mi = ar.master_instance
    #     assert isinstance(mi, rt.models.ratings.RatingsSummary)
    #     qs = super().get_request_queryset(ar, **filter)
    #     qs = qs.filter(master=mi.master, skill=mi.skill)
    #     return qs


class ExamDetail(dd.DetailLayout):
    main = "general #responses more"
    general = dd.Panel("""
    subject group date
    seqno heading
    ratings.ResponsesByExam
    """, _("General"))

    # responses = dd.Panel("""
    # ratings.PupilsAndRatingsByExam
    # """, _("Responses"))

    more = dd.Panel("""
    period id user
    ratings.ChallengesByExam #ratings.RatingsByExam
    """, _("More"))



class Exams(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'ratings.Exam'
    detail_layout = "ratings.ExamDetail"

class ExamsByGroup(Exams):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "group"
    column_names = "subject seqno date heading completion user period *"
    insert_layout = """
    subject
    heading
    """

class ExamsByCourse(Exams):
    required_roles = dd.login_required(PrimaTeacher)
    master = "prima.Course"
    column_names = "heading date seqno completion user period *"
    # default_display_modes = {None: constants.DISPLAY_MODE_SUMMARY}
    extra_display_modes = {constants.DISPLAY_MODE_LIST, constants.DISPLAY_MODE_SUMMARY}
    # card_layout = """
    # seqno date id
    # """
    insert_layout = """
    heading
    date
    """

    @classmethod
    def setup_request(self, ar):
        super().setup_request(ar)
        mi = ar.master_instance
        ar.known_values.update(subject=mi.subject, group=mi.group)

    # @classmethod
    # def get_request_queryset(self, ar, **filter):
    #     mi = ar.master_instance
    #     assert isinstance(mi, rt.models.prima.Course)
    #     qs = super().get_request_queryset(ar, **filter)
    #     return qs.filter(subject=mi.subject, group=mi.group)


# class ExamsByCast(Exams):
#     required_roles = dd.login_required(PrimaTeacher)
#     master = "prima.Cast"
#     column_names = "seqno heading period *"
#
#     @classmethod
#     def setup_request(self, ar):
#         super().setup_request(ar)
#         mi = ar.master_instance
#         ar.known_values.update(subject=mi.subject, group=mi.group, user=mi.user)

    # @classmethod
    # def get_request_queryset(self, ar, **filter):
    #     mi = ar.master_instance
    #     assert isinstance(mi, rt.models.prima.Cast)
    #     qs = super().get_request_queryset(ar, **filter)
    #     if True:
    #         qs = qs.filter(subject=mi.subject, group=mi.group, user=mi.user)
    #     else:
    #         # When there are two teachers for a same subject, they manage tests
    #         # together.
    #         qs = qs.filter(subject=mi.subject, group=mi.group)
    #     return qs


    # @classmethod
    # def get_master_instance(cls, ar, model, pk):
    #     # the master instance of ExamsByGroup must be a group, but since
    #     # we use this on an cast, we get the pk of a cast
    #     assert model is rt.models.prima.Cast
    #     cast = rt.models.prima.Cast.objects.get(pk=pk)
    #     return cast.group

class PupilsAndRatingsByExam(dd.VirtualTable):
    required_roles = dd.login_required(PrimaTeacher)
    master = "ratings.Exam"
    label = _("Pupils and ratings")
    default_display_modes = {None: constants.DISPLAY_MODE_HTML}

    @classmethod
    def table_as_html(cls, ar):
        Enrolment = rt.models.prima.Enrolment
        ExamResponse = rt.models.ratings.ExamResponse
        Challenge = rt.models.ratings.Challenge
        ChallengeRating = rt.models.ratings.ChallengeRating
        exam = ar.master_instance
        # print(f"20241017 cellattrs {ar.renderer.cellattrs}")
        challenges = Challenge.objects.filter(exam=exam)
        if not challenges.exists():
            return _("There are no {challenges} configured for {exam}.").format(
                challenges=Challenge._meta.verbose_name_plural, exam=exam)

        insert_button_attrs = dict(style="text-align: center;")
        def challenge_cell(sar, challenge):
            try:
                rating = ChallengeRating.objects.get(
                    enrolment=sar.master_instance.enrolment, challenge=challenge)
            except ChallengeRating.DoesNotExist:
                btn = sar.gen_insert_button(None, insert_button_attrs,
                        challenge=str(challenge), challengeHidden=challenge.pk)
                if btn is None:
                    return ""
                return btn

            # elems = list(rating.get_rating_buttons(sar))
            # elems.append(sar.obj2html(rating, "🖉"))  # #1f589
            # return E.span(*join_elems(elems, sep=" | "))
            return E.span(*join_elems(rating.get_rating_buttons(sar), sep=" | "))

            # txt = f"{rating.score or NOT_RATED}"
            # kw = dict(action_param_values=dict(score=rating.score))
            # return sar.instance_action_button(
            #     rating.set_rating_action, label=txt, request_kwargs=kw)

        table = xghtml.Table()
        table.attrib.update(ar.renderer.tableattrs)
        headers = [E.td(gettext("Pupil"))]
        cellstyle = "padding:2pt; margin:0pt; text-align:center;"
        # cellstyle = "text-align:center;"
        for chl in challenges:
            # print(f"20241017 {prj.display_color}")
            headers.append(E.td(
                ar.obj2html(chl, str(chl.skill)), style=cellstyle))
        headers.append(E.td(gettext("Remark")))
        table.head.append(E.tr(*headers))
        # for enr in Enrolment.objects.filter(group=exam.cast.group):
        for resp in ExamResponse.objects.filter(exam=exam).order_by("enrolment__pupil__last_name"):
            # sar = ChallengeRatingsByEnrolment.request(parent=ar, master_instance=enr)
            # resp  = rt.models.ratings.ExamResponse.objects.get(exam=exam, enrolment=resp.enrolment)
            sar = RatingsByResponse.request(parent=ar, master_instance=resp)
            # sar._insert_sar = None
            cells = [E.td(ar.obj2html(resp, str(resp.enrolment.pupil)))]
            for chl in challenges:
                cells.append(E.td(challenge_cell(sar, chl), style=cellstyle))
            cells.append(E.td(resp.remark, style=cellstyle))
            table.body.append(E.tr(*cells))

        el = table.as_element()
        # if len(toolbar := ar.plain_toolbar_buttons()):
        #     el = E.div(el, E.p(*toolbar))
        return el


class ExamResponses(dd.Table):
    model = "ratings.ExamResponse"
    detail_layout = """
    exam enrolment
    remark
    ratings.RatingsByResponse
    """

class ResponsesByExam(ExamResponses):
    master_key = "exam"
    # default_display_modes = {None: constants.DISPLAY_MODE_HTML}
    column_names = "enrolment score1 score2 ratings remark *"

    @classmethod
    def get_column_names(cls, ar):
        # Builds columns dynamically by request. Called once per UI handle.
        scores = ' '.join(dd.plugins.ratings.get_score_column_names())
        return f"enrolment {scores} ratings remark *"

    @classmethod
    def override_column_headers(self, ar):
        Challenge = rt.models.ratings.Challenge
        mi = ar.master_instance
        # d = {n: "" for n in dd.plugins.ratings.get_score_column_names()}
        d = dict()
        if mi is not None:
            qs = Challenge.objects.filter(exam=mi).order_by('seqno')
            for chl in qs:
                max_score = format_score(chl.max_score)
                d['score' + str(chl.seqno)] = f"{chl.skill} ({max_score})"
        return d



class ExamResponsesByEnrolment(ExamResponses):
    master_key = "enrolment"
    default_display_modes = { None: constants.DISPLAY_MODE_LIST}
    column_names = "exam ratings remark *"
    insert_layout = """
    exam
    remark
    """


class Challenges(dd.Table):
    required_roles = dd.login_required(PrimaTeacher)
    model = 'ratings.Challenge'
    detail_layout = dd.DetailLayout("""
    exam project_section id
    seqno skill max_score
    ratings.RatingsByChallenge
    """, window_size=(60, 'auto'))

class AllChallenges(Challenges):
    required_roles = dd.login_required(PrimaStaff)

class ChallengesBySkill(Challenges):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "skill"
    row_template = "{row.max_score} in {row.exam_or_section}"

class ChallengesByExam(Challenges):
    required_roles = dd.login_required(PrimaTeacher)
    master_key = "exam"
    column_names = "seqno skill max_score id *"
    # detail_layout = dd.DetailLayout("""
    # exam project_sectionid
    # seqno skill max_score
    # ratings.RatingsByChallenge
    # """, window_size=(60, 'auto'))
    insert_layout = """
    skill
    max_score
    """


class ChallengeRatings(dd.Table):
    required_roles = dd.login_required(PrimaTeacher)
    model = 'ratings.ChallengeRating'
    order_by = ['challenge__exam', 'challenge__project_section', 'challenge__skill']
    # parameters = dict()
    # params_layout = """challenge challenge__exam enrolment"""
    detail_layout = dd.DetailLayout("""
    challenge max_score
    enrolment
    score rating_buttons
    period teacher date_done
    """, window_size=(60, "auto"))

class AllChallengeRatings(ChallengeRatings):
    required_roles = dd.login_required(PrimaStaff)

class ChallengeRatingsByEnrolment(ChallengeRatings):
    master_key = "enrolment"
    default_display_modes = { None: constants.DISPLAY_MODE_LIST}
    insert_layout = """
    challenge
    max_score
    enrolment
    score rating_buttons
    period teacher date_done
    """

    def as_summary_item(self, ar, text=None, **kwargs):
        # must return an ET element
        if text is None:
            exam = self.challenge.exam or self.challenge.project_section
            text = f'{exam}: {self.score} {self.challenge.skill}'
        return super().as_summary_item(ar, text, **kwargs)

class RatingsByResponse(ChallengeRatings):
    master = "ratings.ExamResponse"
    # default_display_modes = { None: constants.DISPLAY_MODE_SUMMARY}
    # column_names = "challenge rating"
    # column_names = "challenge__skill challenge__max_score *"
    column_names = "challenge__skill challenge__max_score score *"
    # column_names = "score *"
    label = _("Ratings by response")
    row_template = '{row.score}/{row.challenge.max_score} {row.challenge.skill}'

    # @classmethod
    # def row_as_summary(cls, ar, obj, text=None, **kwargs):
    #     if text is None:
    #         text = f'{obj.score} {obj.challenge.skill}'
    #     return super().row_as_summary(ar, obj, text, **kwargs)

    # @classmethod
    # def setup_request(self, ar):
    #     super().setup_request(ar)
    #     mi = ar.master_instance
    #     ar.known_values.update(enrolment=resp.enrolment, challenge__exam=resp.exam)

    @classmethod
    def get_request_queryset(cls, ar):
        qs = super().get_request_queryset(ar)
        resp = ar.master_instance
        if resp is None: return qs.none()
        qs = qs.filter(enrolment=resp.enrolment, challenge__exam=resp.exam)
        return qs


class RatingsByChallenge(ChallengeRatings):
    master_key = "challenge"
    label = _("Ratings")
    # default_display_modes = { None: constants.DISPLAY_MODE_LIST}
    column_names = "enrolment score *"

# class RatingsByExam(ChallengeRatings):
#     master_key = "challenge__exam"
#     label = _("Ratings")
#     default_display_modes = { None: constants.DISPLAY_MODE_LIST}
#     column_names = "enrolment challenge__max_score score challenge challenge__skill *"

# class RatingsByCourse(ChallengeRatings):
#     master = "prima.Course"
#     @classmethod
#     def get_request_queryset(self, ar, **filter):
#         mi = ar.master_instance
#         assert isinstance(mi, rt.models.prima.Course)
#         qs = super().get_request_queryset(ar, **filter)
#         return qs.filter(enrolment__group=mi.group, challenge__skill__subject=mi.subject)




class FinalExams(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'ratings.FinalExam'
    column_names = "designation skill max_score *"

class FinalExamRatings(dd.Table):
    required_roles = dd.login_required(PrimaStaff)
    model = 'ratings.FinalExamRating'
    column_names = "enrolment exam score *"
    detail_layout = """
    exam
    enrolment
    score max_score
    teacher period date_done
    """

class FinalRatingsByEnrolment(FinalExamRatings):
    master_key = "enrolment"
    required_roles = dd.login_required(PrimaTeacher)
    column_names = "exam score teacher date_done *"

# class FinalRatingsByCast(FinalExamRatings):
#     master_key = "enrolment"
#     required_roles = dd.login_required(PrimaTeacher)
#     column_names = "exam enrolment score date_done *"


if dd.is_installed("projects"):

    class ChallengesByProjectSection(Challenges):
        required_roles = dd.login_required(PrimaTeacher)
        master_key = "project_section"
        column_names = "seqno skill max_score id *"
        insert_layout = """
        skill
        max_score
        """

    class ChallengeRatingsByProject(ChallengeRatings):
        master = "ratings.Project"
        label = _("Challenge ratings")
        default_display_modes = {
            70: constants.DISPLAY_MODE_HTML,
            None: constants.DISPLAY_MODE_LIST}
        column_names = "max_score score rating_buttons challenge__skill period *"
        group_by = [lambda obj: obj.challenge.project_section]
        obvious_fields = set(['enrolment'])

        @classmethod
        def get_request_queryset(cls, ar):
            qs = super().get_request_queryset(ar)
            prj = ar.master_instance
            if prj is None: return qs.none()
            project_challenges = rt.models.ratings.Challenge.objects.filter(
                project_section__project_template=prj.template)
            qs = qs.filter(challenge__in=project_challenges,
                enrolment=prj.enrolment)
            # if (pv := ar.param_values) is None: return qs
            return qs

        @classmethod
        def before_group_change(cls, gh, obj):
            return format_html("<h2>{}</h2>", obj.challenge.project_section)



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from decimal import Decimal
from django.db import models
from django.conf import settings
from django.utils.html import format_html, mark_safe
from django.utils.text import format_lazy
from django.core.exceptions import ValidationError
from lino.utils.html import E, tostring, join_elems
from lino.utils import SumCollector
from lino.utils.mldbc.mixins import BabelDesignated
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.printing.mixins import CachedPrintable
from lino.modlib.system.choicelists import DisplayColors
from lino.modlib.summaries.mixins import SlaveSummarized
from lino.modlib.checkdata.choicelists import Checker
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _

from .mixins import RatingBase, RatingSummary, RatingCollector
from .mixins import format_percentage, format_score, ScoreField, MaxScoreField, ScoreValue
# from .mixins import RatingSummaryBase
from .choicelists import Predicates, Smilies


# class Exam(UserAuthored, BabelDesignated, Sequenced):   # zeugnisse.Test
class Exam(UserAuthored, Sequenced):   # zeugnisse.Test
    class Meta:
        verbose_name = _("Exam")
        verbose_name_plural = _("Exams")
        abstract = dd.is_abstract_model(__name__, 'Exam')
        ordering = ['group', 'subject', 'seqno']
    group = dd.ForeignKey('prima.Group', related_name='enrolments')
    subject = dd.ForeignKey('prima.Subject')
    heading = dd.CharField(_("Title"), max_length=250)
    # cast = dd.ForeignKey('prima.Cast')
    date = models.DateField(_("Date"), blank=True, null=True)
    period = dd.ForeignKey('periods.StoredPeriod')

    def get_siblings(self):
        return self.__class__.objects.filter(group=self.group, subject=self.subject)

    def on_create(self, ar):
        self.date = dd.today()
        return super().on_create(ar)

    def after_ui_create(self, ar):
        super().after_ui_create(ar)
        self.fix_problems.run_from_ui(ar)

    def __str__(self):
        return self.heading

    def full_clean(self):
        if not self.period_id:
            P = rt.models.periods.StoredPeriod
            date = self.date or dd.today()
            qs = P.objects.filter(
                year=self.group.year,
                start_date__lte=date,
                end_date__gte=date)
            # print("20241030", qs)
            period = qs.first()
            if period is None:
                period = P.get_or_create_from_date(date)
            self.period = period
        if self.period.year != self.group.year:
            raise ValidationError(_("Period {} must be in {}").format(
                self.period.year, self.group.year))
        super().full_clean()

    @dd.displayfield(_("Completion"))
    def completion(self, ar=None):
        if not self.pk:
            return ""
        qs = rt.models.ratings.ChallengeRating.objects.filter(
            challenge__exam=self, enrolment__group=self.group)
        todo = qs.count()
        if todo == 0:
            return NOT_RATED
        done = qs.filter(score__isnull=False).count()
        # return format_score(100 * done / todo) + "%"
        return format_percentage(done, todo)




dd.update_field(Exam, 'user', verbose_name=_("Teacher"))


class Challenge(Sequenced):  # zeugnisse.Leistung
    # Other names: Expectation, ExamQuestion, Achievement, Feat
    class Meta:
        verbose_name = _("Challenge")
        verbose_name_plural = _("Challenges")
        abstract = dd.is_abstract_model(__name__, 'Challenge')
        unique_together = ('exam', 'project_section', 'skill')

    allow_cascaded_delete = ["exam", "project_section"]

    skill = dd.ForeignKey('prima.Skill', blank=True, null=True)
    exam = dd.ForeignKey('ratings.Exam', blank=True, null=True)
    project_section = dd.ForeignKey('projects.ProjectSection', blank=True, null=True)
    max_score = MaxScoreField()

    def get_siblings(self):
        if self.exam_id:
            return self.__class__.objects.filter(exam=self.exam)
        return self.__class__.objects.filter(project_section=self.project_section)

    @dd.displayfield(_("Section"))
    def exam_or_section(self, ar=None):
        if self.exam_id:
            return self.exam.as_summary_item(ar)
        return self.project_section.as_summary_item(ar)
        # return f"{self.project_section.project_template.short_header}:{self.project_section}"

    @dd.chooser()
    def skill_choices(cls, exam, project_section):
        # print(f"20241212 {exam}")
        if exam is not None:
            qs = rt.models.prima.Skill.objects.filter(subject=exam.subject)
        elif project_section is not None:
            qs = rt.models.prima.Skill.objects.all()
        else:
            qs = rt.models.prima.Skill.objects.all()
        return qs

    def __str__(self):
        # return f"{self.project_section or self.exam} ({self.id})"
        # return str(self.project_section or self.exam)
        # return f"{self.max_score} for {self.skill} in {self.project_section or self.exam}"
        # return f"{self.skill} in {self.project_section or self.exam}"
        return _("{skill} /{max_score} in {exam_or_section}").format(
            max_score=format_score(self.max_score), skill=self.skill,
            exam_or_section=self.exam_or_section)

    # def as_summary_item(self, ar, text=None):
    #     if text is None:
    #         text = ""
    #         if ar.is_obvious_field("skill"):
    #             text = f"{self} in {self.exam_or_section}"
    #         elif ar.is_obvious_field("exam") or ar.is_obvious_field("project_section"):
    #             text = f"{self} for {self.skill} in {self.exam_or_section}"
    #         else:
    #             text = str(self.skill)
    #     return super().as_summary_item(ar, text)

    # def as_summary_item(self, ar, text=None):
    #     if ar.is_obvious_field("skill"):
    #         if self.project_section_id is not None:
    #             obj = self.project_section.project_template
    #         if self.exam_id is not None:
    #             obj = self.exam
    #         else:
    #             obj = super()
    #     else:
    #         obj = self.skill or super()
    #     txt = obj.as_summary_item(ar, text)
    #     return format_html("{1} ({0})", self.max_score, tostring(txt))

class ChallengeRating(RatingBase):  # zeugnisse.LeistungsBewertung
    class Meta:
        verbose_name = _("Challenge rating")
        verbose_name_plural = _("Challenge ratings")
        abstract = dd.is_abstract_model(__name__, 'ChallengeRating')
    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"), editable=False)
    challenge = dd.ForeignKey('ratings.Challenge', editable=False)
    allow_cascaded_delete = ['challenge']

    def get_skill(self):
        return self.challenge.skill

    def get_max_score(self):
        return self.challenge.max_score

    def get_rating_type(self):
        return self.challenge.skill.subject.rating_type

    def as_paragraph(self, ar, **kwargs):
        # s = ar.obj2htmls(self)

        # elems = list(self.get_rating_buttons(ar))
        # s = tostring(E.span(*join_elems(elems, sep=" | ")))
        # s = tostring(E.span(*join_elems(self.get_rating_buttons(ar), sep=" | ")))
        s = " | ".join(map(tostring, self.get_rating_buttons(ar)))

        if not ar.is_obvious_field("challenge"):
            s += " " + format_html(_("in {subject}:{skill}"),
                subject=self.challenge.skill.subject,
                skill=self.challenge.skill)
        if not ar.is_obvious_field("enrolment"):
            s += " " + format_html(_("for {pupil}"), pupil=ar.obj2htmls(self.enrolment))
        return mark_safe(s)


class ResponseScoreField(dd.VirtualField):
    editable = True

    def __init__(self, seqno):
        self._seqno = seqno
        t = ScoreField(str(seqno))
        dd.VirtualField.__init__(self, t, None)

    def value_from_object(self, obj, ar):
        try:
            rating = ChallengeRating.objects.get(
                challenge__exam=obj.exam,
                challenge__seqno=self._seqno,
                enrolment=obj.enrolment)
        except ChallengeRating.DoesNotExist:
            return None
        except ChallengeRating.MultipleObjectsReturned:
            qs = ChallengeRating.objects.filter(
                challenge__seqno=self._seqno,
                challenge__exam=obj.exam,
                enrolment=obj.enrolment)
            # print(f"20241212 {qs}")
            return None
        return rating.score

    def set_value_in_object(self, ar, obj, value):
        if ar is None:  # e.g. when called by loaddata
            return
        Enrolment = rt.models.prima.Enrolment
        ExamResponse = rt.models.ratings.ExamResponse
        Challenge = rt.models.ratings.Challenge
        ChallengeRating = rt.models.ratings.ChallengeRating
        assert isinstance(obj, ExamResponse)
        # print(f"20241017 cellattrs {ar.renderer.cellattrs}")
        try:
            challenge = Challenge.objects.get(exam=obj.exam, seqno=self._seqno)
        except Challenge.DoesNotExist:
            ar.warning(_("20241212 Oops"))
            return
        except Challenge.MultipleObjectsReturned:
            ar.warning(_("{} has multiple challenges with seqno {}").format(
                obj, self._seqno))
            return
        rating, created = ChallengeRating.objects.get_or_create(
            enrolment=obj.enrolment, challenge=challenge)
        rating.score = value
        rating.full_clean()
        rating.save()


class ExamResponse(dd.Model):
    class Meta:
        verbose_name = _("Exam response")
        verbose_name_plural = _("Exam responses")
        abstract = dd.is_abstract_model(__name__, 'ExamResponse')
        unique_together = ['enrolment', 'exam']

    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"), editable=False)
    exam = dd.ForeignKey('ratings.Exam', editable=False)
    remark = dd.RichTextField(_("Remark"), blank=True, format="plain")

    allow_cascaded_delete = ["exam"]

    # score1 = ResponseScoreField(1)
    # score2 = ResponseScoreField(2)

    def disabled_fields(self, ar):
        df = super().disabled_fields(ar)
        df |= set(SCORE_FIELDS)
        for o in Challenge.objects.filter(exam=self.exam):
            df.discard("score" + str(o.seqno))
        return df

    @classmethod
    def get_simple_parameters(cls):
        for p in super().get_simple_parameters():
            yield p
        yield 'enrolment'
        yield 'exam'

    @dd.displayfield(_("Ratings"))
    def ratings(self, ar):
        if ar is None:
            return ''
        elems = []

        Enrolment = rt.models.prima.Enrolment
        ExamResponse = rt.models.ratings.ExamResponse
        Challenge = rt.models.ratings.Challenge
        ChallengeRating = rt.models.ratings.ChallengeRating
        exam = ar.master_instance
        # print(f"20241017 cellattrs {ar.renderer.cellattrs}")
        challenges = Challenge.objects.filter(exam=exam)
        if not challenges.exists():
            return _("No {challenges} in {exam}.").format(
                challenges=Challenge._meta.verbose_name_plural, exam=exam)
        sar = ChallengeRatingsByEnrolment.request(
            parent=ar, master_instance=self.enrolment)
        insert_button_attrs = dict(style="text-align: center;")

        def rating_elem(challenge):
            try:
                rating = ChallengeRating.objects.get(
                    enrolment=self.enrolment, challenge=challenge)
            except ChallengeRating.DoesNotExist:
                return f"{NOT_RATED} / {challenge.max_score}"
                # btn = sar.gen_insert_button(None, insert_button_attrs,
                #         challenge=str(challenge), challengeHidden=challenge.pk)
                # if btn is None:
                #     return ""
                # return btn
                # return E.p(txt, btn, align="center")
                # return str("+")
            # txt = f"{rating.score or NOT_RATED}"
            # return sar.obj2html(rating, txt)
            return sar.obj2html(rating)

        for chl in challenges:
            elems.append(rating_elem(chl))
        return E.p(*join_elems(elems, sep=", "))

max_skills_per_exam = dd.get_plugin_setting("ratings", "max_skills_per_exam", 0)

SCORE_FIELDS = []
for i in range(1, max_skills_per_exam+1):
    name = "score" + str(i)
    SCORE_FIELDS.append(name)
    dd.inject_field(ExamResponse, name, ResponseScoreField(i))


class FinalExam(dd.Model):  # zeugnisse.AbschlussTest
    class Meta:
        verbose_name = _("Final exam")
        verbose_name_plural = _("Final exams")
        abstract = dd.is_abstract_model(__name__, 'FinalExam')
        ordering = ['designation']

    designation = dd.CharField(_("Final exam"), max_length=200)
    skill = dd.ForeignKey('prima.Skill', blank=True, null=True)
    max_score = MaxScoreField()

    @classmethod
    def get_simple_parameters(cls):
        for p in super().get_simple_parameters():
            yield p
        yield 'skill'


class FinalExamRating(RatingBase):  # zeugnisse.AbschlussTestBewertung
    class Meta:
        verbose_name = _("Final exam rating")
        verbose_name_plural = _("Final exam ratings")
        abstract = dd.is_abstract_model(__name__, 'FinalExamRating')

    allow_cascaded_delete = ['enrolment', 'exam']

    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"), editable=False)
    exam = dd.ForeignKey('ratings.FinalExam', editable=False)

    def get_skill(self):
        return self.exam.skill

    def get_max_score(self):
        return self.exam.max_score

    def get_rating_type(self):
        return self.exam.skill.subject.rating_type

    @classmethod
    def get_simple_parameters(cls):
        for p in super().get_simple_parameters():
            yield p
        yield 'enrolment'
        yield 'exam'


class RatingsSummary(SlaveSummarized):
    class Meta:
        verbose_name = _("Rating summary")
        verbose_name_plural = _("Rating summaries")
        abstract = dd.is_abstract_model(__name__, 'ChallengeSummary')
        ordering = ('master', 'skill__subject', 'skill', 'enrolment')
    master = dd.ForeignKey('prima.Group')
    skill = dd.ForeignKey('prima.Skill')
    challenge = dd.ForeignKey('ratings.Challenge', blank=True, null=True)
    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"), blank=True, null=True)
    score = dd.DecimalField(_("Score"), max_digits=5, decimal_places=1, blank=True, null=True)
    max_score = dd.DecimalField(_("Max score"), max_digits=5, decimal_places=1, blank=True, null=True)
    delete_them_all = True

    def __str__(self):
        if self.challenge_id is None:
            if self.enrolment_id is None:
                return _("{subject}:{skill} in {group}").format(
                    group=self.master, subject=self.skill.subject, skill=self.skill)
            else:
                return f"{self.enrolment.pupil} {self.score}"
        elif self.enrolment_id is None:
            return f"{self.challenge}"
        return super().__str__()
        # return f"{self.master} {self.skill} {self.challenge} {self.enrolment} -> {self.score}"

    @classmethod
    def update_for_filter(cls, master=None):
        by_enrolment = SumCollector()
        by_challenge = SumCollector()
        by_skill = SumCollector()

        def loop():
            for r in ChallengeRating.objects.filter(
                enrolment__group=master, score__isnull=False
            ):
                yield (r.enrolment, r.challenge, r.score, r.challenge.max_score)
        for enr, challenge, score, max_score in loop():
            v = ScoreValue(score, max_score)
            skill = challenge.skill
            by_enrolment.collect((skill, enr), v)
            by_challenge.collect((skill, challenge), v)
            by_skill.collect(skill, v)

        def create_or_update(v, **kwargs):
            self, created = cls.objects.get_or_create(master=master, **kwargs)
            self.score = round(v._score, 1)
            self.max_score = round(v._max_score, 1)
            self.full_clean()
            self.save()

        for skill, v in by_skill.items():
            create_or_update(v.rebase(100), skill=skill)
        for skill_enr, v in by_enrolment.items():
            skill, enr = skill_enr
            create_or_update(v.rebase(100), enrolment=enr, skill=skill)
        for skill_chl, v in by_challenge.items():
            skill, chl = skill_chl
            create_or_update(v.rebase(100), challenge=chl, skill=skill)

    def get_summary_collectors(self):
        qs = rt.models.ratings.Challenge.objects.filter(ticket=self)
        yield (self.add_from_session, qs)

    def get_scores_to_summarize(self):
        if self.enrolment_id is not None:
            flt.update(enrolment=self.enrolment)
        elif self.challenge_id is not None:
            flt.update(challenge=self.enrolment)
        project_challenges = Challenge.objects.filter(
            project_section__project_template=self.template)
        for r in ChallengeRating.objects.filter(
            challenge__in=project_challenges,
            enrolment=self.enrolment):
            yield (r.score, r.challenge.max_score)
        for prj in rt.models.projects.Project.objects.filter(
            challenge__in=project_challenges,
            enrolment=self.enrolment):
            for gr in GeneralRatings.get_list_items():
                yield (getattr(self, gr.field_name), gr.max_score)


class ChallengeChecker(Checker):
    verbose_name = _("Check for missing or duplicate challenge ratings")
    model = Challenge
    msg_missing = _("No challenge rating for {} pupils.")
    msg_duplicate = _("Duplicate challenge rating(s) for {} pupils.")

    def get_checkdata_problems(self, obj, fix=False):
        Enrolment = rt.models.prima.Enrolment
        missing = []
        duplicate = 0
        if obj.exam_id is not None:
            enrolments = Enrolment.objects.filter(group=obj.exam.group)
        elif obj.project_section_id is not None:
            prjtpl = obj.project_section.project_template
            enrolments = Enrolment.objects.filter(group__grade=prjtpl.grade)
        else:
            raise Exception("20241026 invalid challenge")
        for enr in enrolments:
            qs = ChallengeRating.objects.filter(challenge=obj, enrolment=enr)
            if qs.count() == 0:
                missing.append(ChallengeRating(challenge=obj, enrolment=enr))
            elif qs.count() > 1:
                duplicate += 1
        if duplicate > 0:
            yield (False, format_lazy(self.msg_duplicate, duplicate))
        if len(missing) > 0:
            yield (True, format_lazy(self.msg_missing, len(missing)))
            if fix:
                for row in missing:
                    row.full_clean()
                    row.save()

ChallengeChecker.activate()


class ExamChecker(Checker):
    verbose_name = _("Check for missing or duplicate exam responses")
    model = Exam
    msg_missing = _("No response for {} enrolments.")
    msg_duplicate = _("Duplicate response(s) for {} enrolments.")

    def get_checkdata_problems(self, obj, fix=False):
        Enrolment = rt.models.prima.Enrolment
        ExamResponse = rt.models.ratings.ExamResponse
        missing = []
        duplicate = 0
        enrolments = Enrolment.objects.filter(group=obj.group)
        for enr in enrolments:
            qs = ExamResponse.objects.filter(exam=obj, enrolment=enr)
            if qs.count() == 0:
                missing.append(ExamResponse(exam=obj, enrolment=enr))
            elif qs.count() > 1:
                duplicate += 1
        if duplicate > 0:
            yield (False, format_lazy(self.msg_duplicate, duplicate))
        if len(missing) > 0:
            yield (True, format_lazy(self.msg_missing, len(missing)))
            if fix:
                for row in missing:
                    row.full_clean()
                    row.save()

ExamChecker.activate()


from .ui import *



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from decimal import Decimal
from django.db import models
from django.conf import settings
from django.utils.html import format_html, mark_safe
from django.utils.text import format_lazy
from django.core.exceptions import ValidationError
from lino.utils.html import E, tostring, join_elems
from lino.utils import SumCollector
from lino.utils.mldbc.mixins import BabelDesignated
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.printing.mixins import CachedPrintable
from lino.modlib.system.choicelists import DisplayColors
from lino.modlib.summaries.mixins import SlaveSummarized
from lino.modlib.checkdata.choicelists import Checker
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _

from lino_prima.lib.ratings.mixins import RatingSummary, MaxScoreField, RatingCollector, NOT_RATED
from .choicelists import ProjectStates

USE_GENERAL_RATINGS = True


class ProjectTemplate(BabelDesignated):  # zeignisse.Baustein
    class Meta:
        verbose_name = _("Project template")
        verbose_name_plural = _("Project templates")
        abstract = dd.is_abstract_model(__name__, 'ProjectTemplate')
    short_header = dd.CharField(_("Short header"), max_length=20)
    main_skill = dd.ForeignKey('prima.Skill', verbose_name=_("Main skill"))
    display_color = DisplayColors.field(blank=True, null=True)
    grade = dd.ForeignKey('prima.Grade')

    def as_summary_item(self, ar, text=None):
        if text is None:
            if ar.is_obvious_field("grade"):
                text = f"{self.short_header}"
            else:
                text = f"{self.short_header} ({self.grade})"
        return super().as_summary_item(ar, text)

    def __str__(self):
        return f"{self.short_header} ({self.grade})"
        # return f"{self.enrolment} doing {self.template}"


class ProjectSection(BabelDesignated, Sequenced):  # zeugnisse.BausteinAbschnitt
    class Meta:
        verbose_name = _("Project section")
        verbose_name_plural = _("Project sections")
        abstract = dd.is_abstract_model(__name__, 'ProjectSection')

    project_template = dd.ForeignKey('projects.ProjectTemplate')

    def get_siblings(self):
        return self.__class__.objects.filter(project_template=self.project_template)

    def __str__(self):
        return f"{self.seqno}) {super().__str__()}"

    def as_paragraph(self, ar, **kwargs):
        s = ar.obj2htmls(self)
        if not ar.is_obvious_field("project_template"):
            s += format_html(_(" (in {project_template})"), project_template=ar.obj2htmls(self.project_template))
        # s += "<br/>x" + tostring(self.get_workflow_buttons(ar))
        qs = rt.models.ratings.Challenge.objects.filter(project_section=self)
        s += " ({})".format(", ".join(
            [tostring(e.as_summary_item(ar)) for e in qs.order_by('seqno')]))
        return mark_safe(s)

    def get_ratings_sum(self, enr):
        rc = RatingCollector()
        for r in self.ratings_for_enrolment(enr):
            rc.collect(r.score, r.challenge.max_score,
                text=r.challenge.skill, period=r.period)
        return rc

    def ratings_for_enrolment(self, enr):
          return rt.models.ratings.ChallengeRating.objects.filter(
              enrolment=enr, challenge__project_section=self)


# class ProjectSkill(BabelDesignated):  # zeugnisse.BausteinArbeit
#     class Meta:
#         verbose_name = _("Rating criterion")
#         verbose_name_plural = _("Rating criteria")
#         abstract = dd.is_abstract_model(__name__, 'ProjectSkill')
#
#     max_score = MaxScoreField()


class Project(RatingSummary, CachedPrintable):
    class Meta:
        verbose_name = _("Project")
        verbose_name_plural = _("Projects")
        abstract = dd.is_abstract_model(__name__, 'Project')

    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"))
    template = dd.ForeignKey('projects.ProjectTemplate')
    date_started = models.DateField(_("Started on"), blank=True, null=True)
    remark = dd.TextField(_("Remark"), blank=True)

    def __str__(self):
        if self.template_id is None:
            return super().__str__()
        # return str(self.template)
        return f"{self.enrolment} doing {self.template.short_header}"

    def before_ui_save(self, ar, cw):
        if self.date_started is None:
            self.date_started = dd.today()
        super().before_ui_save(ar, cw)

    def as_summary_item(self, ar, text=None):
        if text is None:
            text = ""
            if not ar.is_obvious_field("template"):
                text += f"{self.template.short_header} "
            if not ar.is_obvious_field("enrolment"):
                text += f"{self.enrolment} "
            text += f"{self.ratings_done}% {self.total_score or NOT_RATED}"
        return super().as_summary_item(ar, text)

    # def get_ratings_todo(self):
    #     project_challenges = Challenge.objects.filter(
    #         project_section__project_template=self.template)
    #     # n = project_challenges.count() + ProjectSkill.objects.count()
    #     n = project_challenges.count() + len(GeneralRatings.get_list_items())
    #     return n

    def get_scores_to_summarize(self):
        project_challenges = rt.models.ratings.Challenge.objects.filter(
            project_section__project_template=self.template)
        for r in rt.models.ratings.ChallengeRating.objects.filter(
            challenge__in=project_challenges,
            enrolment=self.enrolment):
            yield (r.score, r.challenge.max_score)
        for gr in GeneralRatings.get_list_items():
            yield (getattr(self, gr.field_name), gr.max_score)

    def filename_root(self):
        return f"{self._meta.verbose_name}-{self.enrolment.pupil}-{self.template.short_header}-{self.pk}"

    def get_printable_context(self, ar=None, **kwargs):
        kwargs.update(use_bulma_css=True)
        return super().get_printable_context(ar, **kwargs)

    def get_ratings_sum(self):
        rc = RatingCollector()
        qs = rt.models.ratings.ChallengeRating.objects.filter(
          enrolment=self.enrolment, challenge__project_section__project_template=self.template)
        for r in qs:
            rc.collect(r.score, r.challenge.max_score)
        return rc

    def get_general_ratings(self):
        rc = RatingCollector()
        for gr in GeneralRatings.get_list_items():
            rc.collect(getattr(self, gr.field_name), gr.max_score, text=gr.text)
        return rc


# class ProjectRating(RatingBase):  # zeugnisse.BausteinBewertung
#     class Meta:
#         verbose_name = _("Project score")
#         verbose_name_plural = _("Project ratings")
#         abstract = dd.is_abstract_model(__name__, 'ProjectRating')
#
#     allow_cascaded_delete = ['project']
#
#     project = dd.ForeignKey('ratings.Project', blank=True, null=True)
#     project_skill = dd.ForeignKey('ratings.ProjectSkill', blank=True, null=True)
#
#     def __str__(self):
#         return f"{self.project_skill}: {self.score or NOT_RATED}/{self.project_skill.max_score}"


from .ui import *



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""This is the main module of Lino Prima.

.. autosummary::
   :toctree:

   lib
   projects


"""

__version__ = '25.1.0'

srcref_url = 'https://gitlab.com/synodalsoft/prima/blob/master/%s'
docs_url = "https://synodalsoft.gitlab.io/prima"
intersphinx_urls = dict(docs=docs_url)
doc_trees = ['docs']



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""This package contains the :term:`demo projects <demo project>` included with
Lino Prima.

.. autosummary::
   :toctree:

   prima1

"""



1# -*- coding: UTF-8 -*-
# Copyright 2024-2025 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import datetime
from pathlib import Path

combine = datetime.datetime.combine

from commondata.demonames.bel import NameFactory

from django.utils.text import format_lazy
from django.conf import settings
from lino.api import rt, dd, _
from lino.utils import Cycler, i2d, ONE_DAY
from lino.utils import AttrDict
from lino.core.roles import SiteAdmin
from lino.modlib.system.choicelists import DurationUnits
from lino_xl.lib.working.roles import Worker
from lino.utils.quantities import Duration
from lino.utils.mldbc import babel_named as named
from lino.modlib.users.utils import create_user
from lino_xl.lib.working.choicelists import ReportingTypes
# from lino_xl.lib.tickets.choicelists import SiteStates
from lino_xl.lib.tickets.roles import Reporter, TicketsStaff
from lino_xl.lib.tickets.fixtures.demo import TEXTS
from lino.modlib.system.choicelists import DisplayColors

try:
    from lino_book import DEMO_DATA
except ImportError:
    DEMO_DATA = None

TEXTS = Cycler(TEXTS)
LAST = AttrDict()
MAX_RATINGS = Cycler(5, 10, 20, 12, 6, 8, 15, 20, 30)

User = rt.models.users.User
UserTypes = rt.models.users.UserTypes
StoredPeriod = rt.models.periods.StoredPeriod
Group = rt.models.prima.Group
Subject = rt.models.prima.Subject
Role = rt.models.prima.Role
Cast = rt.models.prima.Cast
Grade = rt.models.prima.Grade
ProjectSection = rt.models.projects.ProjectSection
Skill = rt.models.prima.Skill
Course = rt.models.prima.Course
Challenge = rt.models.ratings.Challenge
Enrolment = rt.models.prima.Enrolment
Exam = rt.models.ratings.Exam
CertTemplate = rt.models.cert.CertTemplate
CertSection = rt.models.cert.CertSection
CertElement = rt.models.cert.CertElement
Certificate = rt.models.cert.Certificate

def certtpl(desig, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    return CertTemplate(**kwargs)

def grade(ref, desig, ctpl=None, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    kwargs.update(ref=str(ref))
    if ctpl:
        kwargs.update(cert_template=ctpl)
    return Grade(**kwargs)

def create_role(desig, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    return rt.models.prima.Role(**kwargs)

def create_subject(desig, **kwargs):
    kwargs = dd.str2kw('designation', desig, **kwargs)
    # print(f"20250115 {kwargs}")
    LAST.define('subject', rt.models.prima.Subject(**kwargs))
    return LAST.subject

# def prjskill(desig, max_score=3, **kwargs):
#     kwargs = dd.babelkw('designation', desig, **kwargs)
#     kwargs.update(max_score=max_score)
#     return rt.models.ratings.ProjectSkill(**kwargs)

def create_skill(desig, with_exams=False, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    kwargs.update(subject=LAST.subject, with_exams=with_exams)
    LAST.define('skill', Skill(**kwargs))
    return LAST.skill

def prjtemplate(desig, ref, color, grade, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    kwargs.update(main_skill=LAST.skill)
    kwargs.update(short_header=ref,
        display_color=DisplayColors.get_by_name(color),
        grade=Grade.get_by_ref(grade))
    LAST.define('prjtemplate', rt.models.projects.ProjectTemplate(**kwargs))
    return LAST.prjtemplate

def prjsection(desig, **kwargs):
    kwargs = dd.babelkw('designation', desig, **kwargs)
    kwargs.update(project_template=LAST.prjtemplate)
    LAST.define('prjsection', ProjectSection(**kwargs))
    return LAST.prjsection

def pupil(first, last):
    return User(
        first_name=first, last_name=last,
        # user_type=UserTypes.pupil,
        username=f"{first}.{last}".lower())

def teacher(first, last):
    return User(first_name=first, last_name=last,
        user_type=UserTypes.teacher,
        username=f"{first}.{last}".lower())


def objects():

    yield (ct1 := certtpl("Junior"))
    yield (ct2 := certtpl("Senior"))

    yield grade(1, _("First grade"), ct1)
    yield grade(2, _("Second grade"), ct1)
    yield grade(3, _("Third grade"), ct2)
    yield grade(4, _("Fourth grade"), ct2)
    yield grade(5, _("Fifth grade"), ct2)
    yield grade("6", _("Sixth grade"), ct2)
    yield grade("X", _("Alumni"))

    # yield prjskill(_("Work behaviour"), 3)  # Arbeitsverhalten
    # yield prjskill(_("Time managment"), 3)  # Zeitmanagement
    # yield prjskill(_("Correction"), 3)  # Korrektur
    # yield prjskill(_("Cleanliness"), 3)  # Sauberkeit

    yield (class_teacher := create_role(_("Class teacher")))
    yield create_role(_("Catholic religion"))
    yield create_role(_("Protestant religion"))
    # yield create_role(_("Geography"))
    # yield create_role(_("History"))
    # yield create_role(_("Natural science"))
    # yield create_role(_("Art"))
    yield create_role(_("Music"))
    yield create_role(_("Sport"))
    yield create_role(_("Projects"))

    # Unicode pictograms: https://www.compart.com/en/unicode/block/U+1F300
    sports_options = dict(icon_text="⛹")  # 🏊 🏈 ⚽
    french_options = dict(icon_text="🥐")  # 🗼🥖 🍇
    german_options = dict(icon_text="🥨")
    if dd.is_installed("uploads"):
        Upload = rt.models.uploads.Upload
        # icons_dir = Path(rt.models.prima.__file__).parent / "flaticon"
        if DEMO_DATA:
            icons_dir = DEMO_DATA / "pictograms"
            if not icons_dir.exists():
                raise Exception(f"Oops, {icons_dir} doesn't exist!")
            vol = rt.models.uploads.Volume(root_dir=icons_dir, ref="pictograms")
            yield vol
            for fn in icons_dir.iterdir():
                if fn.suffix == ".png":
                    yield Upload(volume=vol, library_file=fn.name)
            sports_options.update(image_file=Upload.objects.get(library_file="26bd_soccer.png"))
            french_options.update(image_file=Upload.objects.get(library_file="1f347_grapes.png"))
            german_options.update(image_file=Upload.objects.get(library_file="1f34e_apple.png"))


    yield create_subject(_("Science"), advanced=True, icon_text="🔬") # 🧲 🦠
    yield create_skill(_("Geography"))
    yield create_skill(_("History"))
    yield create_skill(_("Natural science"))
    yield create_subject(_("Art"), rating_type='smiley', icon_text="🎨")  # 🖌
    yield create_skill(_("Painting"))
    yield create_skill(_("Textile"))
    yield create_subject(_("Music"), rating_type='predicate', icon_text="🎜")  # 🎼
    yield create_skill(_("Choir"))
    yield create_skill(_("Flute"))
    yield create_subject(_("Sport"), rating_type='predicate', **sports_options)
    yield create_skill(_("Jogging"))
    yield create_skill(_("Swimming"))
    yield create_subject(_("French"), advanced=True, **french_options)
    yield create_skill(_("Reading"))  # Lesen
    yield create_skill(_("Writing"), True)  # Schreiben
    yield create_skill(_("Speaking"), True)  # Sprechen
    yield create_skill(_("Listening"), True)  # Zuhören
    yield create_skill(_("Dealing with texts and media"), True)  # Umgang mit Texten und Medien
    yield create_subject(_("Religion"), advanced=True, icon_text="🕊")
    yield create_skill(_("Knowledge"), True)
    yield create_skill(_("Caring"), True)
    yield create_skill(_("Trusting"), True)
    yield create_subject(_("Mathematics"), advanced=True, icon_text="🖩")  # π
    yield create_skill(_("Arithmetics"), True)
    yield create_skill(_("Stochastics"), True)
    yield create_subject(_("German"), advanced=True, **german_options)
    yield create_skill(_("Reading"))  # Lesen
    yield create_skill(_("Reading comprehension"))  # Leseverständnis
    yield create_skill(_("Writing"), True)  # Schreiben
    yield create_skill(_("Speaking"), True)  # Sprechen
    yield create_skill(_("Listening"), True)  # Zuhören
    yield create_skill(_("Dealing with texts and media"), True)  # Umgang mit Texten und Medien
    yield create_skill(_("Linguistical thinking"), True)  # Über Sprache reflektieren

    yield prjtemplate(_("Telling experiences in an exciting way"), "Erle", "red", "6")  # Erlebnisse spannend erzählen
    yield prjsection(_("Presentation"))
    yield prjtemplate(_("About animals"), "Tier", "red", "6")  # Rund um Tiere
    yield prjsection(_("Presentation"))
    yield prjsection(_("Video"))
    yield prjtemplate(_("Reporting"), "Ber", "red", "6")  # Berichten
    yield prjtemplate(_("Describing"), "Bes", "red", "6")  # Beschreiben
    yield prjtemplate(_("Tell real-life stories "), "Leb", "red", "6")  # Geschichten aus dem Leben erzählen
    yield prjsection(_("Read silently: The storm"))  # Stilles Lesen: Der Schneesturm

    for ct in CertTemplate.objects.all():
        for subj in Subject.objects.all():
            cs = CertSection(cert_template=ct, subject=subj)
            yield cs
            for skl in Skill.objects.filter(subject=subj):
                yield CertElement(cert_section=cs, skill=skl, max_score=MAX_RATINGS.pop())

    nf = NameFactory()

    MALE_FIRST_NAMES = Cycler(nf.get_male_first_names())
    FEMALE_FIRST_NAMES = Cycler(nf.get_female_first_names())
    LAST_NAMES = Cycler(nf.get_last_names())

    for i in range(40):
        yield pupil(MALE_FIRST_NAMES.pop(), LAST_NAMES.pop())
        yield pupil(FEMALE_FIRST_NAMES.pop(), LAST_NAMES.pop())

    for i in range(11):
        yield teacher(MALE_FIRST_NAMES.pop(), LAST_NAMES.pop())
        yield teacher(FEMALE_FIRST_NAMES.pop(), LAST_NAMES.pop())

    # StoredPeriod.get_or_create_from_date(dd.today(-120))
    StoredPeriod.get_or_create_from_date(dd.today())
    PRJSECTIONS = Cycler(ProjectSection.objects.all())
    SKILLS = Cycler(Skill.objects.all())
    SUBJECTS = Cycler(Subject.objects.all())
    ROLES = Cycler(Role.objects.exclude(id=class_teacher.id))
    PERIODS = Cycler(StoredPeriod.objects.all())
    # PUPILS = Cycler(User.objects.filter(user_type=UserTypes.pupil))
    PUPILS = Cycler(User.objects.filter(user_type=None))
    TEACHERS = Cycler(User.objects.filter(user_type=UserTypes.teacher))

    SUBTEA = dict()

    # for i, subj in enumerate(Subject.objects.all()):
    #     SUBTEA[subj] = [TEACHERS.pop() for j in range(1 + i % 3)]


    for a in "AB":
        for g in Grade.objects.filter(cert_template__isnull=False):
            yield (grp := Group(grade=g, designation=str(g)+a, year=PERIODS.pop().year))
            for i in range(18):
                yield Enrolment(pupil=PUPILS.pop(), group=grp)
            for i in range(4):
                yield Course(subject=SUBJECTS.pop(), group=grp)

            # Every group has at least a class_teacher, the other roles are
            # optional:
            yield Cast(role=class_teacher, group=grp, user=TEACHERS.pop())
            for i in range(2):
                yield Cast(role=ROLES.pop(), group=grp, user=TEACHERS.pop())
            # for i, r in enumerate(Role.objects.all()):
            #     if i % 4:
            # for subj, teachers in SUBTEA.items():
            #     yield Course(subject=subj, group=grp)
            #     for tea in teachers:
            #         # yield Cast(user=tea, subject=subj, group=grp)
            #         yield Cast(user=tea, group=grp, role=ROLES.pop())

    for i in range(30):
        yield Challenge(project_section=PRJSECTIONS.pop(), skill=SKILLS.pop(), max_score=MAX_RATINGS.pop())

    CASTS = Cycler(Cast.objects.all())
    for i in range(30):
        cast = CASTS.pop()
        sub = SUBJECTS.pop()
        yield (o := Exam(heading="Test", subject=sub, user=cast.user, group=cast.group))
        SKILLS = Cycler(Skill.objects.filter(subject=sub))
        if len(SKILLS) > 0:
            for j in range(i % 3):
                yield Challenge(exam=o, skill=SKILLS.pop(), max_score=MAX_RATINGS.pop())

    for enr in Enrolment.objects.all():
        for p in StoredPeriod.objects.filter(year=enr.group.year):
            yield Certificate(enrolment=enr, period=p)



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import locale
from decimal import Decimal
from django.utils import timezone
from django.core.exceptions import ValidationError
from django.db import models
from lino.api import dd, rt, _
from lino.mixins import Created
from lino.utils.html import E, join_elems
from lino.modlib.summaries.mixins import Summarized
from .choicelists import Smilies, Predicates, RatingTypes

ZERO = Decimal()
# NOT_RATED = '▨'
# NOT_RATED = '◻' # 25fb white medium square
# NOT_RATED = '□' # 25a1 white square
# NOT_RATED = '▭' # 25ad white rectagle
NOT_RATED = '☐' # 2610 ballot box

# def RatingField(verbose_name=None, **kwargs):
#     if verbose_name is None:
#         verbose_name = _("Rating")
#     kwargs.update(max_length=2, blank=True, verbose_name=verbose_name)
#     return models.CharField(**kwargs)

def ScoreField(verbose_name=None, **kwargs):
    if verbose_name is None:
        verbose_name = _("Score")
    return dd.DecimalField(
        verbose_name, max_digits=4, decimal_places=1,
        blank=True, null=True, **kwargs)

def MaxScoreField(verbose_name=None, **kwargs):
    if verbose_name is None:
        verbose_name = _("Max. score")
    return dd.DecimalField(
        verbose_name, max_digits=4, decimal_places=1, default=10, **kwargs)


class SetScoreValue(dd.Action):
    """Set the score value of this response."""

    label = _("Set score value")
    icon_name = None
    show_in_toolbar = False
    parameters = dict(score=ScoreField(),
        smiley=Smilies.field(blank=True, null=True),
        predicate=Predicates.field(blank=True, null=True))

    params_layout = "score"
    # params_layout = """
    # score
    # smiley
    # predicate
    # """

    def get_action_permission(self, ar, obj, state):
        return not ar.get_user().is_anonymous

    # @dd.chooser(instance_values=True)
    # def rating_choices(cls, rating_type):
    #     # print(f"20241203 {repr(rating_type)}")
    #     if not rating_type:
    #         return []
    #     return rating_type.rating_choicelist.get_list_items()

    def run_from_ui(self, ar, **kw):
        obj = ar.selected_rows[0]
        pv = ar.action_param_values
        for k, v in pv.items():
            setattr(obj, k, v)
        # obj.score = pv.score
        # obj.smiley = pv.smiley
        # obj.predicate = pv.predicate
        obj.before_ui_save(ar, None)
        obj.full_clean()
        obj.save()
        ar.success(refresh=True)


class SetScoreValueDirectly(SetScoreValue):
    no_params_window = True
    params_layout = """
    smiley
    predicate
    """


class Ratable(dd.Model):

    class Meta:
        abstract = True

    score = ScoreField()
    smiley = Smilies.field(blank=True, null=True)
    predicate = Predicates.field(blank=True, null=True)

    set_score_action = SetScoreValue()
    set_score_action_directly = SetScoreValueDirectly()

    def get_skill(self):
        raise NotImplementedError

    def get_max_score(self):
        raise NotImplementedError

    def get_rating_type(self):
        raise NotImplementedError

    def disabled_fields(self, ar):
        rv = super().disabled_fields(ar)
        if self.get_rating_type() is not None:
            rv.add("score")
        return rv

    def full_clean(self):
        super().full_clean()
        rating_type = self.get_rating_type()
        if rating_type is not None and self.score is not None:
            # needed only after csv2db
            setattr(self, rating_type.field_name, str(self.score))
            # if rating_type == RatingTypes.smiley:
            #     self.smiley = str(self.score)
            # elif rating_type == RatingTypes.predicate:
            #     self.predicate = str(self.score)
            # rating = rating_type.rating_choicelist.get_by_value(str(self.score))
            # if rating is None:
            #     raise ValidationError("20241203")
            # self.rating = rating.value
            self.score = None

    # @dd.chooser(instance_values=True)
    # def rating_choices(cls, rating_type):
    #     # print(f"20241203 {repr(rating_type)}")
    #     if not rating_type:
    #         return []
    #     return rating_type.rating_choicelist.get_list_items()

    @dd.virtualfield(RatingTypes.field(blank=True, null=True))
    def rating_type(self, ar):
        return self.get_rating_type()

    @property
    def score_value(self):
        return ScoreValue(self.score, self.get_max_score())

    def __str__(self):
        return str(self.score_value)
        # return f"{self.score or NOT_RATED}/{max_score}"

    def get_rating_buttons(self, ar, detail_link=True):
        if ar is None:
            return

        rating_type = self.get_rating_type()
        if rating_type is None:
            # max_score = format_rating(self.get_max_score())
            # txt = f"{self.score or NOT_RATED}/{max_score}"
            txt = self.score_value.absolute
            kw = dict(action_param_values=dict(score=self.score))
            yield ar.instance_action_button(
                self.set_score_action, label=txt, request_kwargs=kw)
        else:
            found_existing = False
            value = getattr(self, rating_type.field_name)
            for ch in rating_type.rating_choicelist.get_list_items():
                kw = dict(action_param_values={rating_type.field_name: ch.value})
                label = ch.button_text or ch.text
                if value == ch.value:
                    yield E.span(str(label),
                        style="font-size:1.5em")
                        # style="border:5px solid gray; background-color: gray")
                    found_existing = True
                else:
                    yield ar.instance_action_button(
                        self.set_score_action_directly,
                        label=label, request_kwargs=kw, style="font-size:0.7em")
            if value is not None and not found_existing:
                # value is set but was not found in choicelist
                yield str(value) + "!"
        if detail_link:
            # yield ar.obj2html(self, "🖉")  # #1f589
            yield ar.obj2html(self, "🛈")  # #1f6c8

    @dd.displayfield(_("Max. score"), max_length=3)
    def max_score(self, ar=None):
        if self.get_rating_type() is None:
            return str(self.get_max_score())
        return _("N/A")

    @dd.displayfield(_("Rate it"))
    def rating_buttons(self, ar):
        if ar is None:
            return ""
        return E.span(*join_elems(self.get_rating_buttons(
            ar, detail_link=False), sep=", "))


class RatingBase(Ratable):
    class Meta:
        abstract = True
    teacher = dd.ForeignKey('users.User', verbose_name=_("Teacher"), blank=True, null=True, editable=False)
    period = dd.ForeignKey('periods.StoredPeriod', blank=True, null=True)
    # done = models.DateTimeField(_("Done"), editable=False, null=True)
    date_done = models.DateField(_("Done since"), blank=True, null=True)

    def mark_done(self):
        # self.done = timezone.now()
        self.date_done = dd.today()

    def before_ui_save(self, ar, cw):
        if not self.period:
            self.period = rt.models.periods.StoredPeriod.get_or_create_from_date(
                dd.today())
        if not self.teacher:
            self.teacher = ar.get_user()
        if not self.date_done:
            self.mark_done()
        super().before_ui_save(ar, cw)

    @classmethod
    def get_simple_parameters(cls):
        for p in super().get_simple_parameters():
            yield p
        yield 'teacher'
        yield 'period'


def format_score(n):
    if n is None:
        return ""
    if n % 1:
        return locale.str(round(n, 1))
    return str(int(round(n, 0)))
    # return str(round(n, 0))

def format_percentage(done, todo, width=10):
    if not todo:
        return _("No data")
    if done:
        v = int((done / todo) * width) + 1
    else:
        v = 0
    # return "▪" * v + "▫" * (width - v) + " (" + format_score(100 * done / todo) + "%)"
    # return "■" * v + "▤" * (width - v) + " (" + format_score(100 * done / todo) + "%)"
    # return "▬" * v + "▭" * (width - v) + " (" + format_score(100 * done / todo) + "%)"
    return "▮" * v + "▯" * (width - v) + " (" + format_score(100 * done / todo) + "%)"

# from fractions import Fraction

class ScoreValue:

    def __init__(self, score=ZERO, max_score=ZERO, **kwargs):
        self._score = score
        self._max_score = max_score
        # if score > max_score:
        #     raise ValueError("Score greater than max_score")
        for k, v in kwargs.items():
            setattr(self, k, v)

    def rebase(self, max_score):
        if not self._max_score:
            return self
        return self.__class__(self._score * max_score / self._max_score, max_score)

    def __add__(self, other):
        return self.__class__(
            (self._score or ZERO) + (other._score or ZERO),
            (self._max_score or ZERO) + (other._max_score or ZERO))

    def __mul__(self, other):
        return self.__class__(
            (self._score or ZERO) * other,
            (self._max_score or ZERO) * other)

    def __truediv__(self, other):
        return self.__class__(
            (self._score or ZERO) / other,
            (self._max_score or ZERO) / other)

    def __repr__(self):
        return "<ScoreValue({}, {})>".format(self._score, self._max_score)

    def __str__(self):
        return self.absolute

    def __format__(self, fmt):
        if fmt == "%":
            return self.relative
        return self.absolute

    @property
    def absolute(self):
        return self.score + "/" + self.max_score

    @property
    def relative(self):
        if self._score is not None:
            if self._max_score:
                return format_score(100 * self._score / self._max_score) + " %"
        return ""

    @property
    def score(self):
        if self._score is None:
            return NOT_RATED
        return format_score(self._score)

    @property
    def max_score(self):
        return format_score(self._max_score)



class RatingCollector:
    def __init__(self, value=None, done=0, todo=0):
        if value is None:
            value = ScoreValue()
        self.value = value
        self.done = done
        self.todo = todo
        self.ratings = []

    def __str__(self):
        return str(self.value)

    @property
    def score(self):
        return self.value.score

    @property
    def max_score(self):
        return self.value.max_score

    def __add__(self, other):
        return self.__class__(
            self.value+other.value,
            self.done+other.done,
            self.todo+other.todo)

    def collect(self, score, max_score, **kwargs):
        v = ScoreValue(score, max_score, **kwargs)
        self.todo += 1
        self.ratings.append(v)
        if score is None:
            return
        self.value += v
        self.done += 1

    @property
    def percent_done(self):
        if self.todo == 0:
            return _("N/A")
        return format_score(round(100 * self.done / self.todo, 0)) + "%"


# class RatingSummaryBase(dd.Model):
#     class Meta:
#         abstract = True

class RatingSummary(Summarized):

    class Meta:
        abstract = True

    ratings_done = dd.PercentageField(_("Ratings done"), decimal_places=0, blank=True, null=True, editable=False)
    total_score = dd.DecimalField(_("Total score"), max_digits=5, decimal_places=1, blank=True, null=True, editable=False)
    total_max_score = dd.DecimalField(_("Total max score"), max_digits=5, decimal_places=1, blank=True, null=True, editable=False)

    def get_scores_to_summarize(self):
        return []

    def compute_summary_values(self):
        ratings_todo = 0  # self.get_ratings_todo()
        ratings_done = 0
        total_max_score = 0
        total_score = 0
        for score, max_score in self.get_scores_to_summarize():
            ratings_todo += 1
            if score is not None:
                ratings_done += 1
                total_score += score
                total_max_score += max_score

        self.total_score = round(total_score, 1)
        self.total_max_score = round(total_max_score, 1)
        if ratings_todo == 0:
            self.ratings_done = None
        else:
            self.ratings_done = round(ratings_done * 100 / ratings_todo, 0)

        self.full_clean()
        self.save()

    @property
    def total(self):
        return ScoreValue(self.total_score, self.total_max_score)

    def computed_text(self):
        max_score = self.get_max_score()
        if self.total_max_score:
            text = f"{self.total_score or NOT_RATED}/{self.total_max_score}"
            rel_score = round(self.total_score * max_score / self.total_max_score, 1)
            text += f" -> {rel_score}/{max_score}"
        else:
            text = _("N/A")
        return text



class ExamTypes(dd.ChoiceList):
    verbose_name = _("Examination type")
    verbose_name_plural = _("Examination types")

add = ExamTypes.add_item
add("10", _("ProjectTemplates"), "project")   # Bausteine
add("20", _("Tests"), "test")         # Tests
add("30", _("Final tests"), "final")  # Abschlusstests



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
Lino Prima extension of :mod:`lino.modlib.users`.

.. autosummary::
   :toctree:

    fixtures.demo
    fixtures.demo2

"""

from lino.modlib.users import Plugin


class Plugin(Plugin):

    extends_models = ['User']



from atelier.invlib import setup_from_tasks

ns = setup_from_tasks(
    globals(),
    "lino_prima",
    languages="en de fr".split(),
    # tolerate_sphinx_warnings=True,
    locale_dir='lino_prima/lib/prima/locale',
    revision_control_system='git',
    cleanable_files=['docs/api/lino_prima.*'],
    # demo_prep_command="manage.py prep -v3 --noinput --traceback",
    demo_projects=['lino_prima.projects.prima1'])



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

import locale
import traceback
from decimal import Decimal
from django.db import models
from django.conf import settings
from django.utils.html import format_html, mark_safe, escape
from django.utils.text import format_lazy
from django.core.exceptions import ValidationError
from lino.utils.html import E, tostring, join_elems
from lino.utils.mldbc.mixins import BabelDesignated
from lino.modlib.users.mixins import UserAuthored, My
from lino.modlib.printing.mixins import CachedPrintable
from lino.modlib.system.choicelists import DisplayColors
from lino.mixins import Referrable, Sequenced, Created
from lino.api import dd, rt, _

from lino_prima.lib.prima.roles import PrimaStaff, PrimaTeacher
from lino_prima.lib.ratings.mixins import MaxScoreField, RatingSummary, Ratable, ScoreValue
from lino_prima.lib.ratings.mixins import RatingCollector
from lino_prima.lib.ratings.choicelists import Predicates, RatingTypes
from .mixins import CertificateStates



class CertTemplate(BabelDesignated):  # zeugnisse.ZeugnisVorlage
    class Meta:
        verbose_name = _("Certificate template")
        verbose_name_plural = _("Certificate templates")
        abstract = dd.is_abstract_model(__name__, 'CertTemplate')


class CertSection(Sequenced):  # zeugnisse.ZeugnisAbschnitt
    class Meta:
        verbose_name = _("Certificate section template")
        verbose_name_plural = _("Certificate section templates")
        abstract = dd.is_abstract_model(__name__, 'CertificateSection')

    cert_template = dd.ForeignKey('cert.CertTemplate')
    subject = dd.ForeignKey('prima.Subject')
    # advanced = dd.BooleanField(_("Advanced"), default=False)  # Komplexes Benotungsschema
    remark = dd.RichTextField(_("Remark"), blank=True, format="plain")
    max_score = MaxScoreField()
    rating_type = RatingTypes.field(blank=True, null=True)

    def get_siblings(self):
        return self.__class__.objects.filter(cert_template=self.cert_template)

    # def full_clean(self):
    #     if not self.rating_type:
    #         self.rating_type = self.subject.rating_type()
    #     super().full_clean()

    def __str__(self):
        return f"{self.seqno}) {self.subject}"
        # return f"{self.seqno}) {self.subject} in {self.cert_template}"
        # return f"{self.subject} in {self.cert_template}"
        # return f"{self.subject}"

    def as_paragraph(self, ar, **kwargs):
        s = ar.obj2htmls(self)
        if not ar.is_obvious_field("cert_template"):
            s += format_html(_(" (in {cert_template})"), cert_template=ar.obj2htmls(self.cert_template))
        # s += "<br/>x" + tostring(self.get_workflow_buttons(ar))
        qs = rt.models.cert.CertElement.objects.filter(cert_section=self)
        s += " ({})".format(", ".join(
            [e.as_summary_item(ar) for e in qs.order_by('seqno')]))
        return mark_safe(s)

    # def get_section_rating(self, cert):
    #     return SectionRating(self, cert)


class CertElement(Sequenced):  # zeugnisse.ZeugnisElement
    class Meta:
        verbose_name = _("Certificate element template")
        verbose_name_plural = _("Certificate element templates")
        abstract = dd.is_abstract_model(__name__, 'CertElement')
        unique_together = ('cert_section', 'skill')

    cert_section = dd.ForeignKey('cert.CertSection', blank=True, null=True)
    skill = dd.ForeignKey('prima.Skill', blank=True, null=True)
    max_score = MaxScoreField()

    def get_siblings(self):
        return self.__class__.objects.filter(cert_section=self.cert_section)

    def __str__(self):
        return f"{self.seqno}) {self.skill} in {self.cert_section}"


class Certificate(CachedPrintable):   # zeugnisse.Zeugnis
    class Meta:
        verbose_name = _("Certificate")
        verbose_name_plural = _("Certificates")
        abstract = dd.is_abstract_model(__name__, 'Certificate')
        unique_together = ['enrolment', 'period']

    period = dd.ForeignKey('periods.StoredPeriod')
    enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"))
    state = CertificateStates.field(default="default")
    social_skills_comment = dd.TextField(_("Social skills comment"), blank=True)
    final_verdict = dd.TextField(_("Final verdict"), blank=True)

    # print_certificate = PrintCertificate(show_in_toolbar=False)
    # print_certificate_html = PrintCertificate(
    #     show_in_toolbar=False,
    #     build_method='weasy2html',
    #     label=format_lazy("{}{}", _("Certificate"), _(" (HTML)")))

    def __str__(self):
        return f"{self.enrolment} {self.period.nickname}"

    def get_str_words(self, ar):
        # yield str(self._meta.verbose_name)
        if not ar.is_obvious_field("enrolment"):
            yield str(self.enrolment)
        if not ar.is_obvious_field("period"):
            yield self.period.nickname

    def filename_root(self):
        return f"{self._meta.verbose_name}-{self.enrolment.pupil}-{self.period.nickname}-{self.pk}"

    def get_printable_context(self, ar=None, **kwargs):
        kwargs.update(use_bulma_css=True)
        return super().get_printable_context(ar, **kwargs)

    # @dd.displayfield(_("Print"))
    # def print_actions(self, ar):
    #     if ar is None:
    #         return ''
    #     elems = []
    #     elems.append(ar.instance_action_button(self.print_certificate))
    #     elems.append(ar.instance_action_button(self.print_presence_sheet_html))
    #     return E.p(*join_elems(elems, sep=", "))

    # def get_printable_context(self, bm, elem, ar):
    #     ctx = super().get_printable_context(bm, elem, ar)
    #     p = rt.models.periods.StoredPeriod.get_for_date(dd.today())
    #     certificate, created = rt.models.prima.Certificate.get_or_create(
    #         enrolment=self, period=p)
    #     ctx.update(certificate=certificate)


# dd.update_field(Certificate, "user", verbose_name=_("Teacher"))


class SectionResponse(Ratable, RatingSummary):
    # imported from ZeugnisFachKommentar and ZeugnisPrädikatNote
    class Meta:
        verbose_name = _("Certificate section")
        verbose_name_plural = _("Certificate sections")
        abstract = dd.is_abstract_model(__name__, 'SectionResponse')
        app_label = "cert"
        ordering = ["certificate", "section__seqno"]
        # unique_together = ['enrolment', 'period', 'subject']

    certificate = dd.ForeignKey('cert.Certificate', editable=False)
    section = dd.ForeignKey('cert.CertSection', editable=False)
    # enrolment = dd.ForeignKey('prima.Enrolment', verbose_name=_("Pupil"), editable=False)
    # period = dd.ForeignKey('periods.StoredPeriod')
    remark = dd.RichTextField(_("Remark"), blank=True, format="plain")

    def __str__(self):
        return f"{self.section}"

    _numbers = None

    @property
    def numbers(self):
        if self._numbers is None:
            try:
                self._numbers = SectionNumbers(self)
            except Exception as e:
                traceback.print_exc(e)
        return self._numbers

    def get_max_score(self):
        return self.section.max_score

    def get_rating_type(self):
        return self.section.subject.rating_type

    def get_scores_to_summarize(self):
        for r in self.get_challenge_ratings():
            yield (r.score, r.challenge.max_score)
        for r in self.get_final_ratings():
            yield (r.score, r.exam.max_score)
        # for r in self.get_project_ratings():
        #     yield (r.score, r.project_skill.max_score)
        # what about GeneralRatings?

    def get_challenge_ratings(self):
        return rt.models.ratings.ChallengeRating.objects.filter(
            enrolment=self.certificate.enrolment,
            period=self.certificate.period,
            challenge__skill__subject=self.section.subject)

    def get_final_ratings(self):
        if self.get_rating_type() is not None:
            return []
        return rt.models.ratings.FinalExamRating.objects.filter(
            enrolment=self.certificate.enrolment,
            period=self.certificate.period,
            exam__skill__subject=self.section.subject)


# class CertRating(RatingSummary, Ratable):  # zeugnisse.ZeugnisKompetenzNote + zeugnisse.ZeugnisSozialNote

class CertRating(Ratable):  # zeugnisse.ZeugnisKompetenzNote + zeugnisse.ZeugnisSozialNote

    # todo: renamce CertRating to ElementResponse (and its field 'response' to
    # 'section' and 'cert_element' to 'element')

    class Meta:
        verbose_name = _("Certificate element")
        verbose_name_plural = _("Certificate elements")
        abstract = dd.is_abstract_model(__name__, 'CertRating')

    response = dd.ForeignKey('cert.SectionResponse', editable=False, related_name="elements")
    cert_element = dd.ForeignKey('cert.CertElement', editable=False)

    quick_search_fields = 'response__certificate__enrolment__pupil__first_name response__certificate__enrolment__pupil__last_name'

    _numbers = None

    def get_skill(self):
        return self.cert_element.skill

    def get_max_score(self):
        return self.cert_element.max_score

    def get_rating_type(self):
        return self.cert_element.skill.subject.rating_type

    def __str__(self):
        # return f"{self.ratings_done} {self.total_score} / override {self.score}"
        ce = self.cert_element
        # return f"{self.score} / {ce.max_score} in {ce.skill} for {response.certificate.enrolment}"
        return f"Override {self.score} in {ce} for {self.response.certificate.enrolment}"

    def unused_as_paragraph(self, ar, **kwargs):
        if ar.is_obvious_field('certificate'):
            text = str(self.cert_element.skill)
            # text += " (total " + str(self.total_score)
            # text += "%, " + str(self.ratings_done) + "% done)"
            # text += ": " + ar.obj2htmls(self, str(self.score or NOT_RATED))
            text += ": " + ar.obj2htmls(self)
            return mark_safe(text)
        return ar.obj2htmls(self)

    # useful for debugging exceptions during numbers property:
    # def get_numbers(self):
    #     return self.numbers

    @property
    def numbers(self):
        if self._numbers is None:
            self._numbers = ElementNumbers(self)
        return self._numbers

    @dd.displayfield(_("Computed rating"), max_length=3)
    def computed_rating(self, ar=None):
        # max_score = self.get_max_score()
        # score = self.numbers.total.score * 100 / self.numbers.total.max_score
        # reduced = ScoreValue(self.numbers.total * 100 / max_score)
        rebased = self.numbers.total.rebase(self.get_max_score())
        # self.max_score
        return f"{self.numbers.total} = {rebased}"

    @dd.htmlbox(_("Ratings detail"))
    def ratings_report(self, ar):
        s = ""
        if ar is not None:
            lst = [ar.obj2htmls(r) for r in self.get_challenge_ratings()]
            if len(lst):
                s += "<p>"
                s += escape(rt.models.ratings.ChallengeRating._meta.verbose_name_plural) + ": "
                s += " + ".join(lst)
                s += " = " + str(self.numbers.period)
                s += "</p>"
            lst = [ar.obj2htmls(r) for r in self.get_final_ratings()]
            if len(lst):
                s += "<p>"
                s += escape(rt.models.ratings.FinalExamRating._meta.verbose_name_plural) + ": "
                s += " + ".join(lst)
                s += " = " + str(self.numbers.final)
                s += "</p>"
        return mark_safe(s)

    def get_challenge_ratings(self):
        return rt.models.ratings.ChallengeRating.objects.filter(
            enrolment=self.response.certificate.enrolment,
            period=self.response.certificate.period,
            challenge__skill=self.cert_element.skill)

    def get_final_ratings(self):
        return rt.models.ratings.FinalExamRating.objects.filter(
            enrolment=self.response.certificate.enrolment,
            period=self.response.certificate.period,
            exam__skill=self.cert_element.skill)

    def disabled_fields(self, ar):
        df = super().disabled_fields(ar)
        df.add("response__certificate__enrolment")
        df.add("response__certificate__period")
        df.add("cert_element__skill")
        df.add("cert_element__max_score")
        return df

dd.update_field(CertRating, "rating_buttons", verbose_name=_("Override score"))


from lino.modlib.checkdata.choicelists import Checker

class CertificateChecker(Checker):
    verbose_name = _("Check for missing or duplicate certificate ratings")
    model = Certificate
    msg_missing = _("No certificate rating for {} elements.")
    msg_duplicate = _("Duplicate certificate rating(s) for {} elements.")

    def get_checkdata_problems(self, obj, fix=False):
        CertElement = rt.models.cert.CertElement
        CertRating = rt.models.cert.CertRating
        SectionResponse = rt.models.cert.SectionResponse
        missing = []
        duplicate = 0
        tpl = obj.enrolment.group.grade.cert_template
        for ce in CertElement.objects.filter(cert_section__cert_template=tpl):
            qs = CertRating.objects.filter(response__certificate=obj, cert_element=ce)
            if qs.count() == 0:
                missing.append(ce)
                # missing.append(CertRating(response=resp, cert_element=ce))
            elif qs.count() > 1:
                duplicate += 1
        if duplicate > 0:
            yield (False, format_lazy(self.msg_duplicate, duplicate))
        if len(missing) > 0:
            yield (True, format_lazy(self.msg_missing, len(missing)))
            if fix:
                for ce in missing:
                    qs = SectionResponse.objects.filter(
                        certificate=obj, section=ce.cert_section)
                    n = qs.count()
                    if n == 0:
                        resp = SectionResponse(certificate=obj, section=ce.cert_section)
                        resp.full_clean()
                        resp.save()
                    elif n == 1:
                        resp = qs.first()
                    else:
                        print(f"20241128 Oops there are multiple SectionResponse for {obj} {ce.cert_section}")
                        # raise Exception("")
                    row = CertRating(response=resp, cert_element=ce)
                    row.full_clean()
                    row.save()

CertificateChecker.activate()


# class SectionRating:
#     # volatile object used for printing certificates
#     def __init__(self, cs: CertSection, cert):
#         raise Exception("Must replace this by SectionResponse")
#         self.total = RatingCollector()
#
#         for r in rt.models.ratings.ChallengeRating.objects.filter(
#             enrolment=cert.enrolment, period=cert.period,
#             challenge__skill__subject=cs.subject,
#             score__isnull=False):
#             self.total.collect(r.score, r.challenge.max_score)
#
#         skills = CertElement.objects.filter(cert_section=cs).values_list('skill', flat=True)
#         # skills = rt.models.prima.Skill.objects.filter(subject__in=subjects)
#         for r in rt.models.ratings.FinalExamRating.objects.filter(
#             enrolment=cert.enrolment, period=cert.period,
#             exam__skill__in=skills, score__isnull=False):
#             self.total.collect(r.score, r.exam.max_score)
#
#         for r in rt.models.cert.CertRating.objects.filter(
#             cert_element__cert_section=cs,
#             certificate=cert, score__isnull=False):
#             self.total.collect(r.score, r.challenge.max_score)

class SectionNumbers:

    def __init__(self, sr: SectionResponse):
        cert = sr.certificate
        self.average = ScoreValue()
        year = cert.period.year
        self.by_period = [ScoreValue() for p in year.periods.all() ]
        self.elements = []

        for er in sr.elements.all():
            self.elements.append(er)
            for i, value in enumerate(er.numbers.by_period):
                self.by_period[i] += value
                self.average += value


class ElementNumbers:
    # Volatile object used when printing the certificate
    # Attributes:
    # period: challenge ratings during the period of this certificate
    # final: final exam ratings for this certificate

    def __init__(self, ce: CertRating):
        cert = ce.response.certificate
        skill = ce.cert_element.skill
        ms = ce.get_max_score()

        year = cert.period.year
        self.ccoll = {p.id: RatingCollector() for p in year.periods.all()}
        self.fcoll = {p.id: RatingCollector() for p in year.periods.all()}

        for r in rt.models.ratings.ChallengeRating.objects.filter(
            enrolment=cert.enrolment,
            challenge__skill=skill, score__isnull=False):
                c = self.ccoll[r.period_id]
                c.collect(r.score, r.challenge.max_score)

        for r in rt.models.ratings.FinalExamRating.objects.filter(
            enrolment=cert.enrolment,
            exam__skill=skill, score__isnull=False):
                c = self.fcoll[r.period_id]
                c.collect(r.score, r.exam.max_score)

        # for r in rt.models.cert.CertRating.objects.filter(
        #     response__certificate=cert, element=ce, score__isnull=False):
        #     self.final.collect(r.score, r.challenge.max_score)

        self.by_period = []
        self.average = ScoreValue()
        for p in year.periods.all():
            i = self.ccoll[p.id].value.rebase(ms) + self.fcoll[p.id].value.rebase(ms)
            i = i.rebase(ms)
            self.by_period.append(i)
            self.average += i

        self.average = self.average.rebase(ms)

        self.period = self.ccoll[cert.period.id].value
        self.final = self.fcoll[cert.period.id].value

        self.total = self.period.rebase(ms) + self.final.rebase(ms)
        self.total = self.total.rebase(ms)



from .ui import *



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from lino.ad import Plugin, _


class Plugin(Plugin):

    verbose_name = _("Ratings")
    # menu_group = "ratings"
    max_skills_per_exam = 3

    def get_score_column_names(self):
        # names of the dynamic columns in ResponsesByExam
        for i in range(1, self.max_skills_per_exam + 1):
            yield "score" + str(i)

    def setup_config_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        # m.add_action('ratings.ProjectSkills')
        m.add_action('ratings.FinalExams')

    def setup_explorer_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('ratings.Exams')
        m.add_action('ratings.AllChallenges')
        m.add_action('ratings.AllChallengeRatings')
        m.add_action('ratings.FinalExamRatings')
        # m.add_action('ratings.AllProjectRatings')
        m.add_action('ratings.RatingsSummaries')



# Copyright 2015-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from lino.core.roles import UserRole

class PrimaUser(UserRole): pass
class PrimaPupil(PrimaUser): pass
class PrimaTeacher(PrimaUser): pass
class PrimaStaff(PrimaTeacher, PrimaPupil): pass



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""
The main plugin for Lino Prima.

.. autosummary::
   :toctree:

    migrate
    user_types

"""


from lino.ad import Plugin, _


class Plugin(Plugin):

    verbose_name = _("School")
    menu_group = "prima"
    needs_plugins = ['lino.modlib.periods']

    def setup_main_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        # mg = site.plugins.office
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('prima.MyGroups')
        m.add_action('prima.MyCasts')

    def setup_config_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('prima.Groups')
        m.add_action('prima.Subjects')
        m.add_action('prima.Roles')
        m.add_action('prima.Grades')

    def setup_explorer_menu(self, site, user_type, m, ar=None):
        mg = self.get_menu_group()
        m = m.add_menu(mg.app_label, mg.verbose_name)
        m.add_action('prima.Skills')
        # m.add_action('prima.Exams')
        # m.add_action('prima.ProjectSkills')
        # m.add_action('prima.ProjectTemplates')
        # m.add_action('prima.ProjectSections')
        m.add_action('prima.Enrolments')
        m.add_action('prima.Casts')
        m.add_action('prima.Courses')
        # m.add_action('prima.AllChallenges')
        # m.add_action('prima.AllChallengeRatings')
        # m.add_action('prima.AllProjectRatings')

    # def setup_quicklinks(self, tb):
    #     tb.add_action("prima.MyCasts")
    #     tb.add_action("prima.MyGroups")

    def get_dashboard_items(self, user):
        # yield self.site.models.prima.MyCasts
        yield self.site.models.prima.MyGroups



# -*- coding: UTF-8 -*-
# Copyright 2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)

from lino_prima.lib.prima.settings import *


class Site(Site):

    is_demo_site = True
    the_demo_date = 20241009
    languages = "en de"
    # languages = "de"
    default_ui = 'lino_react.react'
    demo_fixtures = ["std", "demo", "demo2", "checkdata"]
    site_locale = "de_BE.UTF-8"

    def get_plugin_configs(self):
        yield ('checkdata', 'responsible_user', 'rolf')
        yield super().get_plugin_configs()

SITE = Site(globals())

DEBUG = True
USE_TZ = True
TIME_ZONE = 'UTC'



# -*- coding: UTF-8 -*-
# Copyright 2016-2024 Rumma & Ko Ltd
# License: GNU Affero General Public License v3 (see file COPYING for details)
"""General demo data for Lino Prima.

"""

from lino.api import rt


def objects():
    """This will be called by the :ref:`dpy` deserializer during
    :cmd:`pm prep` and must yield a list of object instances to
    be saved.

    """
    return []



# -*- coding: UTF-8 -*-
# generated by lino.sphinxcontrib.help_text_builder
# fmt: off
from django.utils.translation import gettext_lazy as _

help_texts = {
    'lino_prima.lib.prima.migrate.Migrator' : _("""The standard migrator for Lino Prima."""),
    'lino_prima.plugins.cert.Certificate.enrolment' : _("""The pupil for whom this certificate has been issued."""),
    'lino_prima.lib.prima.Subject' : _("""Django model user to represent a subject that can potentially be given at this school."""),
    'lino_prima.lib.prima.Subject.advanced' : _("""Whether this subject uses exams for evaluating the pupil’s progress."""),
    'lino_prima.lib.prima.Subject.icon_text' : _("""A single character to use as pictogram in the printed certificate. This is meant to contain a Unicode symbold or pictogram, e.g. from https://www.compart.com/en/unicode/block/U+1F300"""),
}
